/* BibleSwitchboard.c */
#include <PalmOS.h>
#include "BibleSwitchboard_Rsc.h"
#include "BibleSwitchboard.h"

/*********************************************************************
 * Internal Functions
 *********************************************************************/

static Boolean MainFormHandleEvent(EventType * eventP)
{
	Boolean handled = false;
	FormType * frmP;

	UInt16				cardNo;
	LocalID				dbID;
	DmSearchStateType	searchState;

	Err err = errNone;

	switch (eventP->eType)
	{
		case frmOpenEvent:
			frmP = FrmGetActiveForm();
			FrmDrawForm(frmP);
			handled = true;
			break;
		case ctlSelectEvent:
			switch (eventP->data.ctlSelect.controlID){
				case aboutButton:
					FrmAlert(AboutAlert);
					handled = true;
					break;
				case bookNamesButton:
					FrmAlert(bookNamesAlert);
					handled = true;
					break;
				case prefsButton:
					DmGetNextDatabaseByTypeCreator(true, &searchState, sysFileTPanel, prefsCreatorID, true, &cardNo, &dbID);
					if (dbID) {
						SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdPanelCalledFromApp, NULL);
					}
					handled = true;
					break;
			}
			break;
	}
	return handled;
}
static Boolean AppHandleEvent(EventType * eventP) {
	UInt16 formId;
	FormType * frmP;

	if (eventP->eType == frmLoadEvent) {
		/* Load the form resource. */
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);
 		switch (formId) {
			case MainForm:
				FrmSetEventHandler(frmP, MainFormHandleEvent);
				break;
		}
		return true;
	}

	return false;
}
static void AppEventLoop(void) {
	UInt16 error;
	EventType event;

	do {
		/* change timeout if you need periodic nilEvents */
		EvtGetEvent(&event, evtWaitForever);
		if (! SysHandleEvent(&event)) {
			if (! MenuHandleEvent(0, &event, &error)) {
				if (! AppHandleEvent(&event)) {
					FrmDispatchEvent(&event);
				}
			}
		}
	} while (event.eType != appStopEvent);
}
static Err AppStart(void) {
	return errNone;
}
static void AppStop(void) {
	FrmCloseAllForms();
}
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err error=errNone;
	switch (cmd)
	{
		case sysAppLaunchCmdNormalLaunch:
			error = AppStart();
			if (error)
				return error;

			/*
			 * start application by opening the main form
			 * and then entering the main event loop
			 */
			FrmGotoForm(MainForm);
			AppEventLoop();
			AppStop();
			break;
 	}
	return error;
}
