/*
 * bsw.h
 *
 * public header for shared library
 *
 * This wizard-generated code is based on code adapted from the
 * SampleLib project distributed as part of the Palm OS SDK 4.0,
 *
 * Copyright (c) 1994-1999 Palm, Inc. or its subsidiaries.
 * All rights reserved.
 */

#ifndef BSW_H_
#define BSW_H_

/* Palm OS common definitions */
#include <SystemMgr.h>

/* If we're actually compiling the library code, then we need to
 * eliminate the trap glue that would otherwise be generated from
 * this header file in order to prevent compiler errors in CW Pro 2. */
#ifdef BUILDING_BSW
	#define BSW_LIB_TRAP(trapNum)
#else
	#define BSW_LIB_TRAP(trapNum) SYS_TRAP(trapNum)
#endif

/*********************************************************************
 * Type and creator of Sample Library database
 *********************************************************************/

#define		bswCreatorID	'BSwb'
#define		bswTypeID		sysFileTLibrary

/*********************************************************************
 * Internal library name which can be passed to SysLibFind()
 *********************************************************************/

#define		bswName		"BibleSwitchboardLibrary"

/*********************************************************************
 * BibleSwitchLibrary result codes
 * (appErrorClass is reserved for 3rd party apps/libraries.
 * It is defined in SystemMgr.h)
 *********************************************************************/

/* invalid parameter, most likely a NULL pointer */
#define bswErrParam				(appErrorClass |  1)

/* library is not open */
#define bswErrNotOpen			(appErrorClass |  2)

/* returned from bswClose() if the library is still open */
#define bswErrStillOpen			(appErrorClass |  3)

/* The API doesn't know how to hande the reported structure size
 * Either because it has not been set, or because there is a version mismatch
 * The various API routines will try and be backwards compatible,
 * coping with known previous structure sizes.
 */
#define bswErrBadStructureSize	(appErrorClass |  4)

/* returned from bswLaunch() if the Application can not be found */
#define bswErrAppNotInstalled	(appErrorClass |  5)

/* returned from bswLaunch() if the Application has not been run (to set up it's preferences) */
#define bswErrAppNotRun			(appErrorClass |  6)

/* returned from bswLaunch() if the Application version is not supported */
#define bswErrAppVersion		(appErrorClass |  7)

#define bswErrResource			(appErrorClass |  8)

#define bswErrLock				(appErrorClass |  9)

#define bswErrUnknownBook		(appErrorClass | 10)

/*********************************************************************
 * Custom Types Used by the Bible Switchboard
 *********************************************************************/

/*************************************************************************
 * Reference Contexts
 *
 * Space has been left for grouping translations, books, chapters, verses,
 * and for groupings below the verse level (eg. sentance, word, letter)
 * but these are not yet used, and have been created for later expansion of the library.
 * (Apologies to XP programmers)
 *
 * The bswReference structure may be changed to have a context word,
 * and an array of 16 codes.
 *
 * So instead of referencing
 *    bswReference.chapterNumber
 * you would reference
 *    bswReference.numbers[bswContextBitChapter]
 *
 ************************************************************************/

typedef UInt16 bswContext;

#define bswContextEmpty			0x0000

/* allow for grouping translations later */
#define bswContextTG1			0x0001
#define bswContextTG2			0x0002

#define bswContextTranslation	0x0004

/* allow for grouping books (OT, NT, wisdom books, etc.) */
#define bswContextBG1			0x0008
#define bswContextBG2			0x0010

#define bswContextBook			0x0020

/* allow for grouping chapters */
#define bswContextCG1			0x0040
#define bswContextCG2			0x0080

#define bswContextChapter		0x0100

/* allow for grouping verses (themes, paragraphs, verses joined with a single translation) */
#define bswContextVG1			0x0200
#define bswContextVG2			0x0400

#define bswContextVerse			0x0800

/* allow for sub-verse context */
#define bswContextSG1			0x1000
#define bswContextSG2			0x2000
#define bswContextSG3			0x4000
#define bswContextSG4			0x8000

/* Bit Numbers for the above Context constants*/
#define bswContextBitTG1			0
#define bswContextBitTG2			1
#define bswContextBitTranslation	2
#define bswContextBitBG1			3
#define bswContextBitBG2			4
#define bswContextBitBook			5
#define bswContextBitCG1			6
#define bswContextBitCG2			7
#define bswContextBitChapter		8
#define bswContextBitVG1			9
#define bswContextBitVG2			10
#define bswContextBitVerse			11
#define bswContextBitSG1			12
#define bswContextBitSG2			13
#define bswContextBitSG3			14
#define bswContextBitSG4			15

/* some often used combinations of context */
#define bswContextTranslationBook				0x0024
#define bswContextTranslationBookChapter		0x0124
#define bswContextTranslationBookChapterVerse	0x0924
#define bswContextBookChapter					0x0120
#define bswContextBookChapterVerse				0x0920

typedef struct {
	UInt16		num;			/* translation number */
	Char*		code;			/* translation code */
	Char*		name;			/* translation name */
	UInt16		numBooks;		/* total number of books */
	UInt16		numChapters;	/* total number of chapters */
	UInt16		numVerses;		/* total number of verses */
	UInt16*		bookSeq;		/* book number sequence[total number of books] */
	UInt16*		booksChapters;	/* number of chapters[total number of books] */
	UInt16**	chaptersVerses;	/* number of verses[total number of books][number of chapters[total number of books]] */
	MemPtr		ownerData;		/* supplier of this translation can save any structure, eg to store references to dbID or databaseName, or VFS file name */
} bswTranslation;

typedef struct {
	UInt16			structureSize;		/* set to sizeof(bswTranslationList), in order to determine the version */
	UInt16			numTranslations;	/* number of translations */
	UInt32			forAPI;				/* 32 flags indicating which API functions these translations are for */
	bswTranslation*	translationArray;	/* Pointer to array of translations */
	Char**			translationNames;	/* Pointer to array of pointers to translation names, ending in NULL pointer */
	Char**			translationCodes;	/* Pointer to array of pointers to translation codes, ending in NULL pointer */
} bswTranslationList;

#define bswTranslationListStructureSizeV1 20

typedef struct {
	UInt16 structureSize;			/* set to sizeof(bswReference), in order to determine the version */
	/* Primary reference */
	bswTranslation*  translation;	/* Particular bible translation that this is a reference to, or NULL for any */
	UInt16 bookNumber;				/* book number as per http://tirith.mine.nu/jsp/BibleConverterBookNumbers.html */
	UInt16 chapterNumber;			/* chapter numbers start from 1, 0 means "not specified" */
	UInt16 verseNumber;				/* verse numbers start from 1, 0 means "not specified" */
	/* Alternative reference representations, see reference translation APIs */
	Char*  textRef;					/* String representation of the reference, NULL means not translated */
	UInt32 verseRef;				/* verse number counted from the beginning of the bible, 0 means "not specified" */
} bswReference;

#define bswReferenceStructureSizeV1 20

typedef struct {
	UInt16 structureSize;	/* set to sizeof(bswReference), in order to determine the version */
	/* Primary reference */
	UInt16 bookNumber;		/* book number as per http://tirith.mine.nu/jsp/BibleConverterBookNumbers.html */
	UInt16 chapterNumber;	/* chapter numbers start from 1, 0 means "not specified" */
	UInt16 verseNumber;		/* verse numbers start from 1, 0 means "not specified" */
	/* Alternative reference representations, see reference translation APIs */
	Char*  textRef;			/* String representation of the reference, NULL means not translated */
	UInt32 verseRef;		/* verse number counted from the beginning of the bible, 0 means "not specified" */
	/* Versification Scheme and Translation */
	UInt32 schemeRef;		/* If 0 = not specified, then using previously set scheme */
	Char*  translationName;	/* Name of the particular bible translation that this is a reference to, or NULL for any */
} old_bswReference;

//#define bswReferenceStructureSizeV1 24

typedef struct {
	UInt16 structureSize;	/* set to sizeof(bswText), in order to determine the version */
	UInt16 textLength;
	Char*  text;
	bswReference *refP;
} bswText;

#define bswTextStructureSizeV1 12

typedef struct {
	UInt16 structureSize;			/* set to sizeof(bswSearch), in order to determine the version */
	UInt16 flags;					/* To determine the kind of search to perform. */
	bswTranslation*  translation;	/* Particular bible translation to search, or NULL for any */
	Char *searchText;
	MemPtr matchingVerses;
	UInt16 numMatches;
} bswSearch;

#define bswSearchStructureSizeV1 18

typedef struct {
	UInt16 structureSize;	/* set to sizeof(bswCount), in order to determine the version */
	bswContext flagWhatToCount;
	bswReference *withinContextReference;
	bswContext flagWithinWhatContext;
	UInt16 countResult;
} bswCount;

#define bswCountStructureSizeV1 12
#define bswCountUnknown	-1

/* Reference type flags for specifying source and destination when translating/completing references */
typedef UInt16 bswReferenceType;
#define bswStandardReferenceType	 1
#define bswNumericReferenceType		 2
#define bswTextReferenceType		 4
/* Additional text types for deciding how to represent the reference in text */
#define bswTinyTextReferenceType	 8
#define bswShortTextReferenceType	16

/* Applications implementing bswAppLaunchCmdLaunch must also implement bswAppLaunchCmdSelectedReference */
typedef enum {
	bswAppLaunchCmdLaunch = sysAppLaunchCmdCustomBase,
	bswAppLaunchCmdGetReference,
	bswAppLaunchCmdFindText,
	bswAppLaunchCmdSelectReference,
	bswAppLaunchCmdSelectedReference,
	bswAppLaunchCmdGetCount,
	bswAppLaunchCmdGetTranslationList,
	bswAppLaunchCmdFreeTranslationList
} bswLaunchCode;

#define bswAPIMaskLaunch				1
#define bswAPIMaskGetReference			2
#define bswAPIMaskFindText				4
#define bswAPIMaskSelectReference		8
#define bswAPIMaskSelectedReference		16
#define bswAPIMaskGetCount				32
#define bswAPIMaskGetTranslationList	64
#define bswAPIMaskFreeTranslationList	128

/*********************************************************************
 * API Prototypes
 *********************************************************************/

#ifdef __cplusplus
extern "C" {
#endif

/* Standard library open, close, sleep and wake functions */

extern Err bswOpen(UInt16 refNum, UInt32 * clientContextP)
	BSW_LIB_TRAP(sysLibTrapOpen);
				
extern Err bswClose(UInt16 refNum, UInt32 clientContext)
	BSW_LIB_TRAP(sysLibTrapClose);

extern Err bswSleep(UInt16 refNum)
	BSW_LIB_TRAP(sysLibTrapSleep);

extern Err bswWake(UInt16 refNum)
	BSW_LIB_TRAP(sysLibTrapWake);

/*********************************************************************
 * Custom library API functions
 *********************************************************************/

/************************************************************
 *
 *  FUNCTION: bswVersion
 *
 *  DESCRIPTION:	Get our library API version.  The library does not
 *					need to be "opened" to call bswVersion.
 *
 *  PARAMETERS:	refNum		-- library reference number (unused) [left as it was in the sample]
 *				verP		-- pointer to variable for storing the version number
 *
 *  CALLED BY:		Anyone wishing to get our library API version
 *
 *  RETURNS:	errNone		-- no error
 *				bswErrParam	-- if verP is NULL
 *
 *************************************************************/
extern Err bswVersion(UInt16 refNum, UInt32 * verP)
	BSW_LIB_TRAP(sysLibTrapBase + 5);

/************************************************************
 *
 *  FUNCTION: bswHasImplementation
 *
 *  DESCRIPTION:	Find out if the launch code has had an implementation specified, and that the implementor exists.
 *
 *  PARAMETERS:
 *
 *	refNum
 *		Library reference number returned by bswOpen.
 *
 *	launchCode
 *		implementable function to test for the existance of an implementor
 *
 *  CALLED BY:	Anyone wishing to test the capabilities of the current setup.
 *
 *  RETURNS:
 *		bswErrParam if launchCode is NULL.
 *		dmErrCantFind if the specified implementing application can not be found.
 *		errNone if the specified implementing application can be found.
 *
 */
extern Err bswHasImplementation(UInt16 refNum, bswLaunchCode launchCode)
	BSW_LIB_TRAP(sysLibTrapBase + 6);

/************************************************************
 * FUNCTION: bswLaunch
 *
 * DESCRIPTION:
 *
 * Switch to the prefered bible application, viewing the supplied reference.
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by bswOpen.
 *
 * refP
 *      Pointer to reference structure containing the verse to view
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		bswErrParam	if refP is NULL
 *      dmErrCantFind if the specified implementing application can not be found.
 *		memErrNotEnoughSpace if there is not enough space to allocate the launch command parameter block.
 *		errNone
 */
extern Err bswLaunch(UInt16 refNum, bswReference *refP)
	BSW_LIB_TRAP(sysLibTrapBase + 7);

/************************************************************
 * FUNCTION: bswGetReference
 *
 * DESCRIPTION:
 *
 * Retrieve the text for the reference
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by bswOpen.
 *
 * refP
 *		Pointer to structure containing Bible Switchboard Reference
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 */
extern Err bswGetReference(UInt16 refNum, bswText *refP)
	BSW_LIB_TRAP(sysLibTrapBase + 8);

/************************************************************
 * FUNCTION: bswFindText
 *
 * DESCRIPTION:
 *
 * Search the bible for the given text, returning a list of verses (bitmap?)
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by bswOpen.
 *
 * refP
 *		Pointer to search structure
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 */
extern Err bswFindText(UInt16 refNum, bswSearch *refP)
	BSW_LIB_TRAP(sysLibTrapBase + 9);

/************************************************************
 * FUNCTION: bswSelectReference
 *
 * DESCRIPTION:
 *
 * Open the preferred verse selector screen, returning the selected reference
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by bswOpen.
 *
 * refP
 *		Pointer to structure containing Bible Switchboard Reference
 *
 * CALLED BY: User
 *
 * RETURNS:
 *		errNone
 */
extern Err bswSelectReference(UInt16 refNum, bswReference *refP)
	BSW_LIB_TRAP(sysLibTrapBase + 10);

/************************************************************
 * FUNCTION: bswSelectedReference
 *
 * DESCRIPTION:
 *
 * Get the reference currently in use by the preferred bible application
 * 
 * PARAMETERS:
 *
 * refNum
 *		Library reference number returned by bswOpen.
 *
 * refP
 *		Pointer to structure containing Bible Switchboard Reference to fill in
 *
 * CALLED BY: User
 *
 * RETURNS:
 *      dmErrCantFind if the specified implementing application can not be found.
 *		errNone
 */
extern Err bswSelectedReference(UInt16 refNum, bswReference *refP)
	BSW_LIB_TRAP(sysLibTrapBase + 11);

/* Bible Information Query APIs */

extern Err bswGetCount(UInt16 refNum, UInt32 clientContext, bswCount *countP)
	BSW_LIB_TRAP(sysLibTrapBase + 12);

extern Err bswGetTranslationList(UInt16 refNum, UInt32 clientContext, UInt16 whereFrom, bswTranslationList **translationListP)
	BSW_LIB_TRAP(sysLibTrapBase + 13);

extern Err bswFreeTranslationList(UInt16 refNum, UInt32 clientContext, UInt16 whereFrom, bswTranslationList **translationListP)
	BSW_LIB_TRAP(sysLibTrapBase + 14);

/* Book Name APIs */

extern Err bswGetBookID(UInt16 refNum, UInt32 clientContext, UInt16 seqNum, UInt16 *bookIDP)
	BSW_LIB_TRAP(sysLibTrapBase + 15);

extern Err bswGetBooksSeqNum(UInt16 refNum, UInt32 clientContext, UInt16 bookID, UInt16 *seqNumP)
	BSW_LIB_TRAP(sysLibTrapBase + 16);

extern Err bswGetBookNames(UInt16 refNum, UInt32 clientContext, bswReferenceType typeFlag, Char ***allNamesArrayP, MemHandle *namesResourceP)
	BSW_LIB_TRAP(sysLibTrapBase + 17);

extern Err bswGetBooksNames(UInt16 refNum, UInt32 clientContext, UInt16 bookID, UInt16 *numNamesP, Char ***allNamesArrayP, MemHandle *namesResourceP)
	BSW_LIB_TRAP(sysLibTrapBase + 18);

extern Err bswFreeBookNames(UInt16 refNum, UInt32 clientContext, Char ***allNamesArrayP, MemHandle *namesResourceP)
	BSW_LIB_TRAP(sysLibTrapBase + 19);

extern Err bswCompleteRef(UInt16 refNum, UInt32 clientContext, bswReference *refP, bswReferenceType useFlag, bswReferenceType completeFlags)
	BSW_LIB_TRAP(sysLibTrapBase + 20);

/* Utility functions that could be in clients, but are here to ensure they are done correctly, 
  and to save space, as they are often used. */

/* instead use bswGetCount(refNum, clientContext, bswContextBook, bswContextTranslation, refP) */
extern Err bswGetTotalNumBooks(UInt16 refNum, UInt32 clientContext, UInt16 *totNumBooksP)
	BSW_LIB_TRAP(sysLibTrapBase + 21);

/* instead use bswGetCount(refNum, clientContext, bswContextChapter, bswContextTranslationBook, refP) */
extern Err bswGetBooksNumChapters(UInt16 refNum, UInt32 clientContext, UInt16 bookID, UInt16 *booksNumChaptersP)
	BSW_LIB_TRAP(sysLibTrapBase + 22);

/* instead use bswGetCount(refNum, clientContext, bswContextVerse, bswContextTranslationBookChapter, refP) */
extern Err bswGetChaptersNumVerses(UInt16 refNum, UInt32 clientContext, UInt16 bookID, UInt16 chapterNum, UInt16 *chaptersNumVersesP)
	BSW_LIB_TRAP(sysLibTrapBase + 23);

/*
extern Err bswGetSchemeList(UInt16 refNum, SysDBListItemType *schemesP)
	BSW_LIB_TRAP(sysLibTrapBase + 12);

extern Err bswSetScheme(UInt16 refNum, UInt32 clientContext, UInt32 schemeCode)
	BSW_LIB_TRAP(sysLibTrapBase + 13);

extern Err bswGetTotalNumChapters(UInt16 refNum, UInt32 clientContext, UInt16 *totNumChaptersP)
	BSW_LIB_TRAP(sysLibTrapBase + 15);

extern Err bswGetTotalNumVerses(UInt16 refNum, UInt32 clientContext, UInt16 *totNumVersesP)
	BSW_LIB_TRAP(sysLibTrapBase + 16);

extern Err bswGetBooksFirstChapterNum(UInt16 refNum, UInt32 clientContext, UInt16 bookID, UInt16 *booksFirstChapterNumP)
	BSW_LIB_TRAP(sysLibTrapBase + 20);

extern Err bswGetBooksNumVerses(UInt16 refNum, UInt32 clientContext, UInt16 bookID, UInt16 *booksNumVersesP)
	BSW_LIB_TRAP(sysLibTrapBase + 21);

extern Err bswGetBooksFirstVerseNum(UInt16 refNum, UInt32 clientContext, UInt16 bookID, UInt16 *booksFirstVerseNumP)
	BSW_LIB_TRAP(sysLibTrapBase + 22);
extern Err bswGetChaptersFirstVerseNum(UInt16 refNum, UInt32 clientContext, UInt16 bookID, UInt16 chapterNum, UInt16 *firstVerseNumP)
	BSW_LIB_TRAP(sysLibTrapBase + 24);
*/

/* bswCheckReference (with flags to specify whether from the implementor of the bswLaunch, bswGetReference, or bswFindText APIs. A versification may be supplied, but may be ignored) */

/* TODO: Add more library API calls, for search results, history, bookmark management; etc. */

#ifdef __cplusplus
}
#endif

/*
 * FUNCTION: bsw_OpenLibrary
 *
 * DESCRIPTION:
 *
 * User-level call to open the library.  This inline function
 * handles the messy task of finding or loading the library
 * and calling its open function, including handling cleanup
 * if the library could not be opened.
 * 
 * PARAMETERS:
 *
 * refNumP
 *		Pointer to UInt16 variable that will hold the new
 *      library reference number for use in later calls
 *
 * clientContextP
 *		pointer to variable for returning client context.  The client context is
 *		used to maintain client-specific data for multiple client support.  The 
 *		value returned here will be used as a parameter for other library 
 *		functions which require a client context.  
 *
 * CALLED BY: System
 *
 * RETURNS:
 *		errNone
 *		memErrNotEnoughSpace
 *      sysErrLibNotFound
 *      sysErrNoFreeRAM
 *      sysErrNoFreeLibSlots
 *
 * SIDE EFFECTS:
 *		*clientContextP will be set to client context on success, or zero on
 *      error.
 */
 
__inline Err bsw_OpenLibrary(UInt16 *refNumP, UInt32 * clientContextP)
{
	Err error;
	Boolean loaded = false;
	Char bswLocalName[24] = bswName;
	
	/* first try to find the library */
	error = SysLibFind(bswLocalName, refNumP);
	
	/* If not found, load the library instead */
	if (error == sysErrLibNotFound)
	{
		error = SysLibLoad(bswTypeID, bswCreatorID, refNumP);
		loaded = true;
	}
	
	if (error == errNone)
	{
		error = bswOpen(*refNumP, clientContextP);
		if (error != errNone)
		{
			if (loaded)
			{
				SysLibRemove(*refNumP);
			}

			*refNumP = sysInvalidRefNum;
		}
	}
	
	return error;
}

/*
 * FUNCTION: bsw_CloseLibrary
 *
 * DESCRIPTION:	
 *
 * User-level call to closes the shared library.  This handles removal
 * of the library from system if there are no users remaining.
 *
 * PARAMETERS:
 *
 * refNum
 *		Library reference number obtained from bsw_OpenLibrary().
 *
 * clientContext
 *		client context (as returned by the open call)
 *
 * CALLED BY: Whoever wants to close the library
 *
 * RETURNS:
 *		errNone
 *		sysErrParamErr
 */

__inline Err bsw_CloseLibrary(UInt16 refNum, UInt32 clientContext)
{
	Err error;
	
	if (refNum == sysInvalidRefNum)
	{
		return sysErrParamErr;
	}

	error = bswClose(refNum, clientContext);

	if (error == errNone)
	{
		/* no users left, so unload library */
		SysLibRemove(refNum);
	} 
	else if (error == bswErrStillOpen)
	{
		/* don't unload library, but mask "still open" from caller  */
		error = errNone;
	}
	
	return error;
}

#endif /* BSW_H_ */
