/*
 * BiblePrefs.c
 */

#include <PalmOS.h>
#include <SystemResources.h>

#include "BiblePrefs.h"
#include "BiblePrefs_Rsc.h"

static Boolean CalledFromAppMode;
		// True means hide the panel pick list and display a Done button to
		// return to the calling app.

static UInt16 panelCount;
static MemHandle panelIDsH;			// MemHandle to block containing pick list items
static SysDBListItemType *panelIDsP = 0;	// ptr to block containing pick list items

static UInt16 appCount;
static MemHandle appIDsH;			// MemHandle to block containing pick list items
static SysDBListItemType *appIDsP = 0;	// ptr to block containing pick list items

static SysDBListItemType noImplementationsItem;

static UInt32 functionMask;

static BibleSwitchFunctionType *bsf=0;

/* Define the minimum OS version we support */
#define ourMinVersion    sysMakeROMVersion(3,0,0,sysROMStageDevelopment,0)
#define kPalmOS20Version sysMakeROMVersion(2,0,0,sysROMStageDevelopment,0)

static void PanelPickListDrawItem (Int16 itemNum, RectanglePtr bounds, Char **){
	Char * itemText;
	itemText = panelIDsP[itemNum].name;
	WinDrawChars(itemText, StrLen(itemText), bounds->topLeft.x, bounds->topLeft.y);
}

static void CreatePanelPickList (ControlPtr triggerP, ListPtr listP, ListDrawDataFuncPtr func){
	UInt16 currentAppCardNo;
	LocalID currentAppDBID;
	int item;

	SysCreatePanelList(&panelCount, &panelIDsH);
	if (panelCount > 0){
		panelIDsP = MemHandleLock(panelIDsH);
	}
	
	LstSetListChoices(listP, NULL, panelCount);
	LstSetHeight (listP, panelCount);
	LstSetDrawFunction(listP, func);
	SysCurAppDatabase(&currentAppCardNo, &currentAppDBID);
	for (item = 0; item < panelCount; item++) {
		if (panelIDsP[item].dbID == currentAppDBID && panelIDsP[item].cardNo == currentAppCardNo) {
			LstSetSelection(listP, item);
			CtlSetLabel(triggerP, panelIDsP[item].name);
			break;
		}
	}
}

static void AppListDrawFunc (Int16 itemNum, RectanglePtr bounds, Char **applist){
	Char * itemText;
	
	itemText = ((SysDBListItemType*)applist[itemNum])->name;
	WinDrawChars(itemText, StrLen(itemText), bounds->topLeft.x, bounds->topLeft.y);
}

static void CreatePickLists(FormType *frmP){
	int item;
	int func;
	ControlPtr triggerP;
	ListPtr listP;
	UInt32 currentPrefs;
	UInt16 prefsSize;
	UInt16 resIndex;
	MemHandle mh = NULL;
	UInt32 *APIflagsP = NULL;
	UInt32 APIflags = 0;
	UInt32 functionBitMask;
	
	MemSet(&noImplementationsItem,sizeof(SysDBListItemType),0);
	SysStringByIndex(appMessages,noImplementations,noImplementationsItem.name,dmDBNameLength);
	
	SysCreateDataBaseList(sysFileTApplication, 0, &appCount, &appIDsH, true);
	if (appCount > 0) {
		appIDsP = MemHandleLock(appIDsH);
		for (item = 0; item < appCount; item++) {
			DmOpenRef dbRef = DmOpenDatabase(appIDsP[item].cardNo,appIDsP[item].dbID,dmModeReadOnly);
			if (dbRef) {
				resIndex = DmFindResourceType(dbRef,appFileCreator,0);
				if (resIndex!=-1){
					mh = DmGetResourceIndex(dbRef,resIndex);
					if (mh) {
						APIflagsP = (UInt32*)MemHandleLock(mh);
						if (APIflagsP) {
							APIflags = *APIflagsP;
							MemHandleUnlock(mh);
						}
						DmReleaseResource(mh);
						for (func=0,functionBitMask=1;func<MAXFUNCTIONBITS;func++,functionBitMask<<=1){
							if (functionMask & functionBitMask) {
								if (APIflags & functionBitMask) {
									if (bsf[func].numItems<MAXLISTMATCHES){
										bsf[func].dbItems[bsf[func].numItems]=&appIDsP[item];
										bsf[func].numItems++;
									}
								}
							}
						}
					}
				}
				DmCloseDatabase(dbRef);
			}
		}
	}
	for (func=0,functionBitMask=1;func<MAXFUNCTIONBITS;func++,functionBitMask<<=1){
		if (functionMask & functionBitMask) {
			triggerP = FrmGetObjectPtr (frmP, FrmGetObjectIndex (frmP, 2001+func));
			listP = FrmGetObjectPtr (frmP, FrmGetObjectIndex (frmP, 3001+func));
			prefsSize=4;
			currentPrefs=0;
			PrefGetAppPreferences(appFileCreator,func+1,&currentPrefs,&prefsSize,true);
			if (bsf[func].numItems==0) {
				bsf[func].dbItems[0]=&noImplementationsItem;
				bsf[func].numItems=1;
			}
			LstSetListChoices(listP, (Char**)bsf[func].dbItems, bsf[func].numItems);
			LstSetHeight (listP, bsf[func].numItems);
			LstSetDrawFunction(listP, AppListDrawFunc);
			if (currentPrefs==0) {
				currentPrefs = bsf[func].dbItems[0]->creator;
				PrefSetAppPreferences(appFileCreator,func+1,currentPrefsVersion,&currentPrefs,4,true);
			}
			for (item = 0; item < bsf[func].numItems; item++){
				if (bsf[func].dbItems[item]->creator == currentPrefs){
					LstSetSelection(listP, item);
					CtlSetLabel(triggerP, bsf[func].dbItems[item]->name);
					break;
				}
			}
		}
	}
}

static void MainFormInit(FormType *frmP){
	UInt16 doneButtonIndex;
	UInt32 currentPrefs;
	UInt16 prefsSize;
	// Configure the form differently if CalledFromAppMode
	if (CalledFromAppMode) {
		FrmHideObject(frmP, FrmGetObjectIndex (frmP, bibleSwitchPanelPopupTrigger));
		FrmShowObject(frmP, FrmGetObjectIndex (frmP, bibleSwitchPanelNameLabel));
		
		// The Done button is supposed to be available for the user to return
		// to the calling app.
		doneButtonIndex = FrmGetObjectIndex (frmP, bibleSwitchDoneButton);
		FrmShowObject(frmP, doneButtonIndex);
	}
	prefsSize=4;
	currentPrefs=0;
	PrefGetAppPreferences(appFileCreator,0,&currentPrefs,&prefsSize,true);
	FrmSetControlValue(frmP,FrmGetObjectIndex (frmP, 5000),(currentPrefs & 1));
}

static Boolean MainFormHandleEvent(EventType * eventP){
	Boolean handled = false;
	FormType * frmP;
	EventType		newEvent;
	UInt16	 		currentAppCardNo;
	LocalID			currentAppDBID;
	UInt32 currentPrefs;
	UInt16 prefsSize;

	switch (eventP->eType){
		case frmOpenEvent:
			frmP = FrmGetActiveForm();
			CreatePickLists(frmP);
			FrmDrawForm(frmP);
			MainFormInit(frmP);
			if (!CalledFromAppMode){
				CreatePanelPickList(FrmGetObjectPtr (frmP, FrmGetObjectIndex (frmP, bibleSwitchPanelPopupTrigger)),FrmGetObjectPtr (frmP, FrmGetObjectIndex (frmP, bibleSwitchPanelList)),PanelPickListDrawItem);
			}
			handled = true;
			break;
			
		case ctlSelectEvent:{
			switch (eventP->data.ctlSelect.controlID){
				case 5000:
					prefsSize=4;
					currentPrefs=0;
					PrefGetAppPreferences(appFileCreator,0,&currentPrefs,&prefsSize,true);
					if (eventP->data.ctlSelect.on) {
						currentPrefs |= 1;
					} else {
						currentPrefs &= 0xFFFE;
					}
					PrefSetAppPreferences(appFileCreator,0,currentPrefsVersion,&currentPrefs,4,true);
					handled = true;
					break;
				case bibleSwitchDoneButton:
					// By simply exiting we return to the prior which called us.
					MemSet(&newEvent, sizeof(newEvent), 0);	// Initialize new event structure
					newEvent.eType = appStopEvent;
					EvtAddEventToQueue(&newEvent);
					handled = true;
					break;
			}
			break;
		}
		case popSelectEvent:{
			switch (eventP->data.popSelect.listID){
				case bibleSwitchPanelList:{
					// Switch to the panel if it isn't this one
					SysCurAppDatabase(&currentAppCardNo, &currentAppDBID);
					if (panelIDsP[eventP->data.popSelect.selection].dbID != currentAppDBID ||
						 panelIDsP[eventP->data.popSelect.selection].cardNo != currentAppCardNo){
						SysUIAppSwitch(panelIDsP[eventP->data.popSelect.selection].cardNo, 
							panelIDsP[eventP->data.popSelect.selection].dbID, 
							sysAppLaunchCmdNormalLaunch, 0);
					}
					
					handled = true;
					break;
				}
				default:{
					int func;
					for(func=0;func<MAXFUNCTIONBITS;func++){
						if (eventP->data.popSelect.listID==3001+func){
							PrefSetAppPreferences(appFileCreator,func+1,currentPrefsVersion,&(bsf[func].dbItems[eventP->data.popSelect.selection]->creator),4,true);
							break;
						}
					}
				}
			}
		}
	}
	return handled;
}

static Boolean AppHandleEvent(EventType * eventP){
	UInt16 formId;
	FormType * frmP;

	if (eventP->eType == frmLoadEvent)
	{
		/* Load the form resource. */
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);

		/* 
		 * Set the event handler for the form.  The handler of the
		 * currently active form is called by FrmHandleEvent each
		 * time is receives an event. 
		 */
		switch (formId)
		{
			case MainForm:
				FrmSetEventHandler(frmP, MainFormHandleEvent);
				break;

		}
		return true;
	}

	return false;
}

static void AppEventLoop(void){
	UInt16 error;
	EventType event;

	do 
	{
		/* change timeout if you need periodic nilEvents */
		EvtGetEvent(&event, evtWaitForever);

		if (! SysHandleEvent(&event))
		{
			if (! MenuHandleEvent(0, &event, &error))
			{
				if (! AppHandleEvent(&event))
				{
					FrmDispatchEvent(&event);
				}
			}
		}
	} while (event.eType != appStopEvent);
}

static Err AppStart(void){
	functionMask=ResLoadConstant(0);
	if (functionMask) {
		bsf=MemPtrNew(sizeof(BibleSwitchFunctionType)*MAXFUNCTIONBITS);
		if (bsf){
			MemSet(bsf,sizeof(BibleSwitchFunctionType)*MAXFUNCTIONBITS,0);
		} else {
			functionMask=0;
			return -1;
		}
	}

	return errNone;
}

static void AppStop(){
	FrmCloseAllForms ();
	if (panelIDsP)	MemPtrFree(panelIDsP);
	if (appIDsP)	MemPtrFree(appIDsP);
	if (bsf)		MemPtrFree(bsf);
}

UInt32 PilotMain(UInt16 cmd, void * cmdPBP, UInt16 launchFlags){
	Err					error;
	UInt16				cardNo;
	LocalID				dbID;
	DmSearchStateType	searchState;
	UInt32 				result;
	
	// We don't support any action codes
	if (cmd != sysAppLaunchCmdNormalLaunch &&
		cmd != sysAppLaunchCmdPanelCalledFromApp &&
		cmd != sysAppLaunchCmdReturnFromPanel) {
		return sysErrParamErr;
	}

	CalledFromAppMode = (cmd == sysAppLaunchCmdPanelCalledFromApp);
	
	if (!CalledFromAppMode)
	{
		PrefActivePanelParamsType prefs;
		
		// set this panel as the active one next time the prefs app is opened
		prefs.activePanel = appFileCreator;
		DmGetNextDatabaseByTypeCreator(true, &searchState, sysFileTApplication,	sysFileCPreferences, true,	&cardNo, &dbID);
		if (dbID) {
			SysAppLaunch(cardNo, dbID, 0, prefAppLaunchCmdSetActivePanel, &prefs, &result);
		}
	}
	error = AppStart();
	FrmGotoForm(MainForm);
	if (!error) AppEventLoop();
	AppStop();
	return 0;
}
