/*
 * BiblePlusInterface.c
 *
 * main file for BiblePlus Interface
 *
 * This wizard-generated code is based on code adapted from the
 * stationery files distributed as part of the Palm OS SDK 4.0.
 *
 * Copyright (c) 1999-2000 Palm, Inc. or its subsidiaries.
 * All rights reserved.
 */
 
#include <PalmOS.h>
#include <VFSMgr.h>

#include "bsw.h"
#include "BiblePlusInterface_Rsc.h"
#include "SharedInterface_Rsc.h"
#include "BiblePlusInterface.h"

UInt32 PilotMain(UInt16 launchCode, void * cmdPBP, UInt16 launchFlags){
	#pragma unused(launchFlags)
	Err err = errNone;
	switch (launchCode){
		case bswAppLaunchCmdLaunch:
			err = Launch((bswReference*)cmdPBP);
			break;
		case bswAppLaunchCmdGetReference:
			err = GetReference((bswText *)cmdPBP);
			break;
		case bswAppLaunchCmdSelectedReference:
			err = SelectedReference((bswReference *)cmdPBP);
			break;
		case bswAppLaunchCmdGetCount:
			err = GetCount((bswCount *)cmdPBP);
			break;
		case bswAppLaunchCmdGetTranslationList:
			err = GetTranslationList((bswTranslationList *)cmdPBP);
			break;
		case bswAppLaunchCmdFreeTranslationList:
			err = FreeTranslationList((bswTranslationList *)cmdPBP);
			break;
	}
	return err;
} 

static Err Launch(bswReference * refP) {
	UInt16 newBook;
	UInt16 newChapter;
	UInt16 newVerse;
	DmOpenRef dbRef = 0;
	void* pRecord;
	Err err=errNone;
	MemHandle hRecord;
	DmSearchStateType stateInfo;
	UInt16 cardNo;
	LocalID dbID,theDatabaseId;
	Boolean bLaunch=false;
	UInt16 viewalert;
	OptionsBuffer myOptionsBuffer;

	if (refP == NULL) return bswErrParam;
	if( refP->structureSize != bswReferenceStructureSizeV1 ) return bswErrBadStructureSize;

	newBook=refP->bookNumber;
	newChapter=refP->chapterNumber;
	newVerse=refP->verseNumber;

	err = DmGetNextDatabaseByTypeCreator(true, &stateInfo, 'appl', bibleAppCreator, true, &cardNo, &dbID);
	if( errNone != err ){
		viewalert=ViewInstallAlert;
	} else {
		theDatabaseId = DmFindDatabase(0, bibleAppPrefsFile);
		if( theDatabaseId == 0 )
		{
			viewalert=ViewRunAlert;
		} else {
			dbRef = DmOpenDatabase(0, theDatabaseId, dmModeReadWrite);
			if(  dbRef ){
				hRecord = DmGetRecord(dbRef,0);
				if( hRecord != 0 ){
					pRecord = MemHandleLock(hRecord);
					if ( pRecord ){		
						if (MemPtrSize(pRecord)>=SizeOfOptionsRecord) {
							if (refP->translation) {
								myOptionsBuffer.TranslationNumber = ((TranslationData*)(refP->translation->ownerData))->ordinal;
							}
							myOptionsBuffer.CurrentBook = newBook;
							if (newChapter==0) newChapter=1;
							myOptionsBuffer.CurrentChapter = newChapter;
							if (newVerse==0) newVerse=1;
							myOptionsBuffer.CurrentVerse = newVerse;
							myOptionsBuffer.Offset = 0;
							if (refP->translation) {
								err = DmWrite(pRecord, VerseOffset, &myOptionsBuffer, sizeof(OptionsBuffer));
							} else {
								err = DmWrite(pRecord, VerseOffset+2, &(myOptionsBuffer.CurrentBook), sizeof(OptionsBuffer)-2);
							}
							if (errNone==err){
								bLaunch=true;
							}
						} else {
							viewalert=ViewVersionAlert;
						}
						MemHandleUnlock(hRecord);
					}
					err = DmReleaseRecord(dbRef, 0, true);
				}
				DmCloseDatabase(dbRef);
			}
		}
	}
	if (bLaunch) {
		err=SysUIAppSwitch(cardNo,dbID,sysAppLaunchCmdNormalLaunch,NULL);
	} else {
		FrmAlert(viewalert);
	}
	return err;
}

static Err SelectedReference(bswReference * refP){
	UInt16 newBook=0;
	UInt16 newChapter=0;
	UInt16 newVerse=0;
	LocalID theDatabaseId;
	DmOpenRef dbRef;
	MemHandle hRecord;
	MemPtr pRecord;

	if (refP == NULL) return bswErrParam;
	if( refP->structureSize != bswReferenceStructureSizeV1 ) return bswErrBadStructureSize;

	theDatabaseId = DmFindDatabase(0, bibleAppPrefsFile);
	if(theDatabaseId) {
		dbRef = DmOpenDatabase(0, theDatabaseId, dmModeReadOnly);
		if(dbRef) {
			hRecord = DmGetRecord(dbRef, 0);
			if(hRecord) {
				if (MemHandleSize(hRecord)>=SizeOfOptionsRecord) {
					pRecord = MemHandleLock(hRecord);
					if ( pRecord ) {
						newBook = ((UInt16*)pRecord)[(VerseOffset>>1)+1];
						newChapter = ((UInt16*)pRecord)[(VerseOffset>>1)+2];
						newVerse = ((UInt16*)pRecord)[(VerseOffset>>1)+3];
						MemHandleUnlock(hRecord);
					}
				}
				DmReleaseRecord(dbRef, 0, false);
			}
			DmCloseDatabase(dbRef);
		}
	}
	refP->bookNumber=newBook;
	refP->chapterNumber=newChapter;
	refP->verseNumber=newVerse;
	return errNone;
}

static Err GetCount(bswCount * countP){

	if (countP == NULL) return bswErrParam;
	if( countP->structureSize != bswCountStructureSizeV1 ) return bswErrBadStructureSize;

	countP->countResult=(UInt16)bswCountUnknown;
	switch (countP->flagWhatToCount){
		case bswContextTranslation:
			countP->countResult=1;
			break;
		case bswContextBook:
			countP->countResult=66;
			break;
		case bswContextChapter:
			countP->countResult=150;
			break;
		case bswContextVerse:
			countP->countResult=150;
			break;
	}
	return errNone;
}

static Err openDatabase(DatabaseReference *dbP){
	Err err = errNone;
	if (dbP == NULL)
		return bswErrParam;
	if (dbP->isVFS) {
		UInt16 volRefNum;
		UInt32 volIterator = vfsIteratorStart;
		dbP->dbRefVFS = NULL;
		while ((volIterator != vfsIteratorStop) && (dbP->dbRefVFS == NULL)){
			err = VFSVolumeEnumerate(&volRefNum, &volIterator);
			if (err == errNone) {
				err = VFSFileOpen(volRefNum, dbP->fileName, vfsModeRead, &(dbP->dbRefVFS));
			}
		}
	} else {
		dbP->dbRefRAM = DmOpenDatabase(dbP->cardNo, dbP->dbID, dmModeReadOnly);
	}
	return err;
}

static Err closeDatabase(DatabaseReference *dbP){
	Err err = errNone;
	if (dbP == NULL)
		return bswErrParam;
	if (dbP->isVFS) {
		err = VFSFileClose(dbP->dbRefVFS);
	} else {
		err = DmCloseDatabase(dbP->dbRefRAM);
	}
	return err;
}

static MemHandle readRecord(DatabaseReference *dbP, UInt16 recIndex){
	Err err = errNone;
	MemHandle mh = NULL;
	if (dbP == NULL)
		return NULL;
	if (dbP->isVFS) {
		err = VFSFileDBGetRecord(dbP->dbRefVFS, recIndex, &mh, NULL, NULL);
	} else {
		mh = DmQueryRecord(dbP->dbRefRAM,recIndex);;
	}
	return mh;
}

static void finishedWithRecord(DatabaseReference *dbP, MemHandle mh){
	Err err = errNone;
	if (dbP == NULL)
		return;
	if (dbP->isVFS) {
		MemHandleFree(mh);
	}
}

static Err GetReference(bswText * textP) {

	Err err;
	int i;

	DmSearchStateType stateInfo;
	
	DatabaseReference db;
	MemHandle hHeaderRecord;
	BibleHeaderRecord* pHeaderRecord;

	UInt16 indexRecordIndex;
	MemHandle hIndexRecord;
	UInt16* pIndexRecord;

	UInt16 chaptersFirstVerse;
	UInt32 versesOffset = 0;
	UInt16 numChapters;
	UInt16 *lastVerseNumber;
	UInt32 *chapterOffset;
	UInt16 *verseOffset;

	UInt32 myVerseRef = 0;
	UInt16 indexOffset = 0;

	UInt16 textOffset = 0;
	UInt16 textOffsetRecord = 0;
	UInt16 tokenLength = 0;
	UInt16 tokenLengthCopied = 0;
	UInt16 tokenLengthRecord = 0;

	UInt16 textRecordIndex;
	MemHandle hTextRecord;
	tokenReference* pTextRecord;
	tokenReference* pTextRecordCopy;

	UInt32 lengthLexicon;
	UInt16 numLexiconRecords;
	MemHandle hLexiconIndexRecord;
	WordInfoType* pLexiconIndexRecord;
	MemHandle* hLexiconRecord;
	UInt8** pLexiconRecord;

	Boolean needSpace;

	if ( textP == NULL ) return bswErrParam;
	if ( textP->refP == NULL ) return bswErrParam;
	if ( textP->structureSize != bswTextStructureSizeV1 ) return bswErrBadStructureSize;

	//locate bible database
	if ( textP->refP->translation ) {
		db.cardNo = ((TranslationData*)(textP->refP->translation->ownerData))->cardNo;
		db.dbID = ((TranslationData*)(textP->refP->translation->ownerData))->dbID;
		db.fileName = ((TranslationData*)(textP->refP->translation->ownerData))->fullName;
	} else {
		//No translation specified... currently just find first one in RAM
		//TODO: use VFS if none in RAM. Possibly enumerate translation list, and use above code.
		err = DmGetNextDatabaseByTypeCreator(true, &stateInfo, dmxVersionType, bibleCreator, true, &(db.cardNo), &(db.dbID));
		if( errNone != err ) {
			return bswErrResource;
		}
	}
	db.isVFS = (UInt16)(db.dbID == NULL);
	err = openDatabase(&db);
	if( err == errNone ) {
		//read header record
		hHeaderRecord = readRecord(&db,0);
		if( hHeaderRecord != 0 ) {
			pHeaderRecord = MemHandleLock(hHeaderRecord);
			if ( pHeaderRecord ) {
				//find the book records
				indexRecordIndex = 0;
				while ((pHeaderRecord->bookInfo[indexRecordIndex].bookNum != textP->refP->bookNumber) && (indexRecordIndex < pHeaderRecord->totalBooks)) {
					indexRecordIndex++;
				}
				if (indexRecordIndex < pHeaderRecord->totalBooks) {
					//read the book index record
					hIndexRecord = readRecord(&db, pHeaderRecord->bookInfo[indexRecordIndex].bookIndex);
					if ( hIndexRecord != 0 ) {
						pIndexRecord = MemHandleLock(hIndexRecord);
						if ( pIndexRecord ) {
							//find the verse offset
							numChapters = pIndexRecord[0];
							lastVerseNumber = &(pIndexRecord[1]);
							chapterOffset = (UInt32*) &(lastVerseNumber[numChapters]);
							verseOffset = (UInt16*) &(chapterOffset[numChapters]);
							if ( textP->refP->chapterNumber == 0 ) textP->refP->chapterNumber = 1;
							if ( textP->refP->verseNumber == 0 ) textP->refP->verseNumber = 1;
							if ( textP->refP->chapterNumber > numChapters ) textP->refP->chapterNumber = numChapters;
							if ( textP->refP->chapterNumber > 1 ) {
								if ( textP->refP->verseNumber > (lastVerseNumber[textP->refP->chapterNumber - 1] - lastVerseNumber[textP->refP->chapterNumber - 2]) ) textP->refP->verseNumber = lastVerseNumber[textP->refP->chapterNumber] - lastVerseNumber[textP->refP->chapterNumber - 1];
							} else {
								if ( textP->refP->verseNumber > lastVerseNumber[textP->refP->chapterNumber - 1] ) textP->refP->verseNumber = lastVerseNumber[textP->refP->chapterNumber - 1];
							}
							chaptersFirstVerse = 0;
							if (textP->refP->chapterNumber > 1) {
								chaptersFirstVerse = lastVerseNumber[textP->refP->chapterNumber - 2];
							}
							versesOffset = chapterOffset[textP->refP->chapterNumber - 1];
							if (textP->refP->verseNumber > 1) {
								versesOffset += verseOffset[chaptersFirstVerse + textP->refP->verseNumber - 2];
							}
							//find the verse length
							if (textP->refP->verseNumber > 1) { //true
								tokenLength = verseOffset[chaptersFirstVerse + textP->refP->verseNumber - 1] - verseOffset[chaptersFirstVerse + textP->refP->verseNumber - 2];
							} else {
								tokenLength = verseOffset[chaptersFirstVerse];
							}
							//read the verse record
							pTextRecordCopy = MemPtrNew(tokenLength * sizeof(tokenReference));
							if (pTextRecordCopy) {
								tokenLengthCopied = 0;
								if (pHeaderRecord->versionAttrib & infByteNotShifted) {
									//find the verse record
									textRecordIndex = pHeaderRecord->bookInfo[indexRecordIndex].bookIndex + 1 + ((UInt16) (versesOffset >> 11));
									textOffset = (UInt16) (versesOffset & 2047);
									textOffsetRecord = textOffset;
									while (tokenLengthCopied < tokenLength) {
										tokenLengthRecord = tokenLength - tokenLengthCopied;
										if (textOffsetRecord + tokenLengthRecord > 2048) {
											tokenLengthRecord = 2048 - textOffsetRecord;
										}
										hTextRecord = readRecord(&db, textRecordIndex);
										if ( hTextRecord != 0 ) {
											pTextRecord = MemHandleLock(hTextRecord);
											if ( pTextRecord ) {
												MemMove(&(pTextRecordCopy[tokenLengthCopied]),&(pTextRecord[textOffsetRecord]),(Int32) (tokenLengthRecord * sizeof(tokenReference)));
												MemHandleUnlock(hTextRecord);
											}
											finishedWithRecord(&db,hTextRecord);
										}
										tokenLengthCopied += tokenLengthRecord;
										if (tokenLengthCopied < tokenLength) {
											textRecordIndex++;
											textOffsetRecord = 0;
										}
									}
								} else {
									UInt8 *pCompressedTextRecord = 0;
									UInt32 compStart;
									UInt16 decShift;
									UInt32 decValueBuffer[3];
									UInt32 value;

									//find the verse record
									compStart = versesOffset * 7 / 4;
									textRecordIndex = pHeaderRecord->bookInfo[indexRecordIndex].bookIndex + 1 + ((UInt16) (compStart >> 12));
									textOffset = (UInt16) (compStart & 4095);
									textOffsetRecord = textOffset;
									hTextRecord = readRecord(&db, textRecordIndex);
									if ( hTextRecord != 0 ) {
										pCompressedTextRecord = MemHandleLock(hTextRecord);
										if ( pCompressedTextRecord ) {
											switch (versesOffset * 7 % 4) {
												case 0:
													decShift = 10; // aaaa aaaa : aaaa aabb : 0000 0000
													break;
												case 1:
													decShift = 8; // ccdd dddd : dddd dddd : 0000 0000
													decValueBuffer[2] = pCompressedTextRecord[textOffsetRecord];
													textOffsetRecord++;
													break;
												case 2:
													decShift = 6; // bbbb cccc : cccc cccc : ccdd dddd
													decValueBuffer[2] = pCompressedTextRecord[textOffsetRecord];
													textOffsetRecord++;
													break;
												case 3:
													decShift = 4; // aaaa aabb : bbbb bbbb : bbbb cccc
													decValueBuffer[1] = pCompressedTextRecord[textOffsetRecord];
													textOffsetRecord++;
													break;
											}
										}
									}
									while (tokenLengthCopied < tokenLength){
										if (textOffsetRecord >= 4096) {
											textOffsetRecord = 0;
											textRecordIndex++;
											MemHandleUnlock(hTextRecord);
											finishedWithRecord(&db,hTextRecord);
											pCompressedTextRecord = 0;
											hTextRecord = readRecord(&db, textRecordIndex);
											if ( hTextRecord != 0 ) {
												pCompressedTextRecord = MemHandleLock(hTextRecord);
											}
										}
										if (pCompressedTextRecord) {
											switch(decShift){
												case 10:
													decValueBuffer[0] = pCompressedTextRecord[textOffsetRecord];
													textOffsetRecord++;
													if (textOffsetRecord >= 4096) {
														textOffsetRecord = 0;
														textRecordIndex++;
														MemHandleUnlock(hTextRecord);
														finishedWithRecord(&db,hTextRecord);
														pCompressedTextRecord = 0;
														hTextRecord = readRecord(&db, textRecordIndex);
														if ( hTextRecord != 0 ) {
															pCompressedTextRecord = MemHandleLock(hTextRecord);
														}
													}
													if (pCompressedTextRecord) {
														decValueBuffer[1] = pCompressedTextRecord[textOffsetRecord];
														textOffsetRecord++;
													} else {
														decValueBuffer[1] = 0;
													}
													decValueBuffer[2] = 0;
													break;
												case 4:
													decValueBuffer[0] = decValueBuffer[1];
													decValueBuffer[1] = pCompressedTextRecord[textOffsetRecord];
													textOffsetRecord++;
													if (textOffsetRecord >= 4096) {
														textOffsetRecord = 0;
														textRecordIndex++;
														MemHandleUnlock(hTextRecord);
														finishedWithRecord(&db,hTextRecord);
														pCompressedTextRecord = 0;
														hTextRecord = readRecord(&db, textRecordIndex);
														if ( hTextRecord != 0 ) {
															pCompressedTextRecord = MemHandleLock(hTextRecord);
														}
													}
													if (pCompressedTextRecord) {
														decValueBuffer[2] = pCompressedTextRecord[textOffsetRecord];
														textOffsetRecord++;
													} else {
														decValueBuffer[2] = 0;
													}
													break;
												case 6:
													decValueBuffer[0] = decValueBuffer[2];
													decValueBuffer[1] = pCompressedTextRecord[textOffsetRecord];
													textOffsetRecord++;
													if (textOffsetRecord >= 4096) {
														textOffsetRecord = 0;
														textRecordIndex++;
														MemHandleUnlock(hTextRecord);
														finishedWithRecord(&db,hTextRecord);
														pCompressedTextRecord = 0;
														hTextRecord = readRecord(&db, textRecordIndex);
														if ( hTextRecord != 0 ) {
															pCompressedTextRecord = MemHandleLock(hTextRecord);
														}
													}
													if (pCompressedTextRecord) {
														decValueBuffer[2] = pCompressedTextRecord[textOffsetRecord];
														textOffsetRecord++;
													} else {
														decValueBuffer[2] = 0;
													}
													break;
												case 8:
													decValueBuffer[0] = decValueBuffer[2];
													decValueBuffer[1] = pCompressedTextRecord[textOffsetRecord];
													textOffsetRecord++;
													decValueBuffer[2] = 0;
													break;
											}

											value = decValueBuffer[0] << 16 | decValueBuffer[1] << 8 | decValueBuffer[2];
											value = value >> decShift;
											value = value & 0x3FFF;

											switch(decShift){
												case 10:
													decShift = 4;
													break;
												case 4:
													decShift = 6;
													break;
												case 6:
													decShift = 8;
													break;
												case 8:
													decShift = 10;
													break;
											}

											// Convert 3FFC to 3FFF back to FFFC to FFFF
											if (value > 0x3FF0){
												value |= 0xC000;
											}

											pTextRecordCopy[tokenLengthCopied] = (tokenReference) value;
										}
										tokenLengthCopied++;
									}
									if (pCompressedTextRecord) {
										MemHandleUnlock(hTextRecord);
										finishedWithRecord(&db,hTextRecord);
									}
								}
								//read the lexicon data
								hLexiconIndexRecord = readRecord(&db, pHeaderRecord->wordIndex);
								if ( hLexiconIndexRecord != 0 ) {
									pLexiconIndexRecord = MemHandleLock(hLexiconIndexRecord);
									if ( pLexiconIndexRecord ) {
										lengthLexicon = 0;
										for ( i = 0; i < pLexiconIndexRecord->totalIndexes; i++) {
											lengthLexicon += ((UInt32)(pLexiconIndexRecord->wordIndex[i].wordLength)) * ((UInt32)(pLexiconIndexRecord->wordIndex[i].totalWord));
										}
										numLexiconRecords = (UInt16) ((lengthLexicon + 4095L) / 4096L);
										hLexiconRecord = MemPtrNew(numLexiconRecords * sizeof(MemHandle));
										if ( hLexiconRecord != 0 ) {
											MemSet(hLexiconRecord, (Int32) (numLexiconRecords *  sizeof(MemHandle)), 0);
											pLexiconRecord = MemPtrNew(numLexiconRecords * sizeof(char*));
											if ( pLexiconRecord != 0 ) {
												MemSet(pLexiconRecord, (Int32) (numLexiconRecords * sizeof(char*)), 0);
												for (i = 0; i < numLexiconRecords; i++) {
													hLexiconRecord[i] = readRecord(&db, pHeaderRecord->wordIndex + 1 + i);
													if ( hLexiconRecord[i] != 0 ) {
														pLexiconRecord[i] = MemHandleLock(hLexiconRecord[i]);
													}
												}
												//process the token list, determining the string length needed
												//(presume a space is needed between EVERY token)
												textP->textLength = 0;
												for (i = 0; i < tokenLength; i++) {
													textP->textLength += tokenStringLength(pLexiconIndexRecord, pLexiconRecord, pTextRecordCopy[i]);
												}
												//allocate space for the string
												textP->text = MemPtrNew(textP->textLength + 1);
												if (textP->text) {
													MemSet(textP->text, textP->textLength + 1, 0);
													//process the token list filling in the string
													textP->textLength = 0;
													needSpace = false;
													for (i = 0; i < tokenLength; i++) {
														addTokenToString(pHeaderRecord, pLexiconIndexRecord, pLexiconRecord, pTextRecordCopy[i], &(textP->textLength), textP->text, &needSpace);
													}
													MemPtrSetOwner (textP->text, 0);
												} else {
													textP->textLength = 0;
												}
												//free up everything
												for (i = 0; i < numLexiconRecords; i++) {
													if ( pLexiconRecord[i] != 0 ) {
														MemHandleUnlock(hLexiconRecord[i]);
														finishedWithRecord(&db,hLexiconRecord[i]);
													}
												}
												MemPtrFree(pLexiconRecord);
											}
											MemPtrFree(hLexiconRecord);
										}
										MemHandleUnlock(hLexiconIndexRecord);
									}
									finishedWithRecord(&db,hLexiconIndexRecord);
								}
								MemPtrFree(pTextRecordCopy);
							}
							MemHandleUnlock(hIndexRecord);
						}
						finishedWithRecord(&db,hIndexRecord);
					}
				}
				MemHandleUnlock(hHeaderRecord);
			}
			finishedWithRecord(&db,hHeaderRecord);
		}
		closeDatabase(&db);
	}

	//return the string

	return err;
}

void getWordIndexLen(WordInfoType* pLexiconIndexRecord, tokenReference token, UInt32 *decWordIndexP, UInt16 *decWordLenP, Boolean *decCompressedP) {
	UInt16 i;
	UInt32 totalByteAcc = 0;
	UInt16 relNum;
	if (token > 0) {
		relNum = token - 1;
		for (i=0; i<pLexiconIndexRecord->totalIndexes; i++){
			if (relNum < pLexiconIndexRecord->wordIndex[i].totalWord){
				*decWordLenP = pLexiconIndexRecord->wordIndex[i].wordLength;
				*decWordIndexP = totalByteAcc + relNum * pLexiconIndexRecord->wordIndex[i].wordLength;
				*decCompressedP = pLexiconIndexRecord->wordIndex[i].compressed;
				i = pLexiconIndexRecord->totalIndexes;
			} else {
				totalByteAcc += pLexiconIndexRecord->wordIndex[i].totalWord * pLexiconIndexRecord->wordIndex[i].wordLength;
				relNum -= pLexiconIndexRecord->wordIndex[i].totalWord;
			}
		}
	} else {
		*decWordLenP = 0;
		*decWordIndexP = 0;
		*decCompressedP = false;
	}
}

UInt16 tokenStringLength(WordInfoType* pLexiconIndexRecord, UInt8** pLexiconRecord, tokenReference token) {
	UInt16 stringLength=0;
	UInt32 decWordIndex; 
	UInt16 decWordLen;
	Boolean decCompressed;
	tokenReference subToken;
	UInt16 i;

	//TODO: no text before verse tag?
	switch (token){
	case 0 :
		break;
	case bookTextType :
		break;
	case chapTextType :
		break;
	case descTextType :
		break;
	case versTextType :
		break;
	default:
		getWordIndexLen(pLexiconIndexRecord, token, &decWordIndex, &decWordLen, &decCompressed);
		if (decCompressed) {
			for (i = 0; i < decWordLen / 2; i++) {
				subToken = (tokenReference) pLexiconRecord[decWordIndex/4096][decWordIndex & 4095] * 256;
				decWordIndex++;
				subToken += pLexiconRecord[decWordIndex/4096][decWordIndex & 4095];
				decWordIndex++;
				stringLength += tokenStringLength(pLexiconIndexRecord, pLexiconRecord, subToken);
			}
		} else {
			stringLength += 1; // assume space required before this word
			stringLength += decWordLen;
		}
	}
	return stringLength;
}

void addTokenToString(BibleHeaderRecord* pHeaderRecord, WordInfoType* pLexiconIndexRecord, UInt8** pLexiconRecord, tokenReference token, UInt16* pos, Char* text, Boolean* needSpace) {
	UInt16 stringLength=0;
	UInt32 decWordIndex; 
	UInt16 decWordLen;
	Boolean decCompressed;
	tokenReference subToken;
	UInt16 i;
	char c;

	//TODO: no text before verse tag?
	switch (token){
	case 0 :
		break;
	case bookTextType :
		break;
	case chapTextType :
		break;
	case descTextType :
		break;
	case versTextType :
		break;
	default:
		getWordIndexLen(pLexiconIndexRecord, token, &decWordIndex, &decWordLen, &decCompressed);
		if (decCompressed) {
			for (i = 0; i < decWordLen / 2; i++) {
				subToken = (tokenReference) pLexiconRecord[decWordIndex/4096][decWordIndex & 4095] * 256;
				decWordIndex++;
				subToken += pLexiconRecord[decWordIndex/4096][decWordIndex & 4095];
				decWordIndex++;
				addTokenToString(pHeaderRecord, pLexiconIndexRecord, pLexiconRecord, subToken, pos, text, needSpace);
			}
		} else {
			for (i=0; i < decWordLen; i++) {
				text[*pos + stringLength + i] = (char) pLexiconRecord[decWordIndex/4096][decWordIndex & 4095];
				decWordIndex++;
			}
			c = text[*pos + stringLength];
			switch (c) {
			case '.': case ',': case '?': 
			case '!': case ';': case ':':
				text[*pos + stringLength + i] = pHeaderRecord->sepChar;
				stringLength++;
				*needSpace = false;
				break;
			default:
				if (*needSpace) {
					for (i = decWordLen; i > 0; i--) {
						text[*pos + stringLength + i] = text[*pos + stringLength + i - 1];
					}
					text[*pos + stringLength] = pHeaderRecord->sepChar;
					stringLength++;
				}
				*needSpace = true;
				break;
			}
			stringLength += decWordLen;
			*pos += stringLength;
		}
	}
}

static void copyTranslationInfo(bswTranslation *mpTranslationP, BibleHeaderRecord *hdr) {
	UInt16 i;
	mpTranslationP->code = MemPtrNew(StrLen(hdr->versionName)+1);	/* translation code */
	MemPtrSetOwner(mpTranslationP->code,0);
	StrCopy(mpTranslationP->code,hdr->versionName);
	mpTranslationP->name = MemPtrNew(StrLen(hdr->versionInfo)+1);	/* translation name */
	MemPtrSetOwner(mpTranslationP->name,0);
	StrCopy(mpTranslationP->name,hdr->versionInfo);
	mpTranslationP->numBooks = hdr->totalBooks;						/* total number of books */
	mpTranslationP->numChapters = 0;								/* total number of chapters */
	mpTranslationP->numVerses = 0;									/* total number of verses */
	mpTranslationP->bookSeq = (UInt16*) MemPtrNew(sizeof(UInt16) * hdr->totalBooks);
	MemPtrSetOwner(mpTranslationP->bookSeq,0);
	for (i=0;i<hdr->totalBooks;i++){
		mpTranslationP->bookSeq[i] = hdr->bookInfo[i].bookNum;		/* book number sequence[total number of books] */
	}
	mpTranslationP->booksChapters = NULL;							/* number of chapters[total number of books] */
	mpTranslationP->chaptersVerses = NULL;							/* number of verses[total number of books][number of chapters[total number of books]] */
}

static UInt16 countcardTranslations(const char *path) {
	Err err=errNone;
	UInt32 fType, fCreator;
	UInt16 totalRecord;
	UInt16 translationsCard;
	UInt32 volIterator = vfsIteratorStart;
	UInt32 dirIterator;
	UInt16 volRefNum;
	FileInfoType info;
	FileRef dirRef;
	FileRef fileRef;
	Char fileName[256];
	Char fullName[256];

	translationsCard = 0;
	if (!FtrGet(sysFileCVFSMgr, vfsFtrIDVersion, &dirIterator)){
		while (volIterator != vfsIteratorStop) {
			err = VFSVolumeEnumerate(&volRefNum, &volIterator);
			if (err == errNone) {
				err = VFSFileOpen(volRefNum, path, vfsModeRead, &dirRef);
				if (err != errNone)
					continue;
				info.nameP = fileName; // point to local buffer
				info.nameBufLen = 256;
				dirIterator = vfsIteratorStart;
		
				while (dirIterator != vfsIteratorStop) {
					// Get the next file
					err = VFSDirEntryEnumerate(dirRef, &dirIterator, &info);
					if (err != errNone)
						continue;
					StrPrintF(fullName, "%s%s", path, fileName);
					if (VFSFileOpen(volRefNum, fullName, vfsModeRead, &fileRef))
						continue;
					if (VFSFileDBInfo(fileRef, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &fType, &fCreator, &totalRecord) || fType != dmxVersionType || fCreator != bibleCreator || !totalRecord) {
						VFSFileClose(fileRef);
						continue;
					}
					//Why leave the file open?
					VFSFileClose(fileRef);
		
					translationsCard++;
				}
				VFSFileClose(dirRef);
			}
		}
	}
	return translationsCard;
}

static Err GetTranslationList(bswTranslationList *tListP) {
	UInt16 i;
	DmSearchStateType state;
	UInt16 cardNo;
	LocalID dbID;
	Err err=errNone;
	bswTranslation * mpTranslation = NULL;
	Char* * mpTranslationNames = NULL;
	Char* * mpTranslationCodes = NULL;
	TranslationData * mpTranslationData = NULL;
	UInt16 translationCount;
	DmOpenRef dRef;
	MemHandle mhR0;
	MemPtr hdr;

	UInt32 fType, fCreator;
	UInt16 totalRecord;
	UInt32 volIterator = vfsIteratorStart;
	UInt32 dirIterator;
	UInt16 volRefNum;
	FileInfoType info;
	FileRef dirRef;
	FileRef fileRef;
	Char fileName[256];
	Char fullName[256];
	UInt16 translationsCard = 0;
	char *path;

	UInt32 bswPrefs;
	UInt16 bswPrefsSize;

	if ( tListP == NULL ) return bswErrParam;
	if ( tListP->structureSize != bswTranslationListStructureSizeV1 ) return bswErrBadStructureSize;

	tListP->numTranslations = 0;
	tListP->translationArray = NULL;
	tListP->translationNames = NULL;
	tListP->translationCodes = NULL;

	// later code could allocate each as it is discovered, and chain them, then allocate the array and copy them sorting as you go.
	// for now, just count them, then allocate space, then fill the space
	err = DmGetNextDatabaseByTypeCreator(true, &state, dmxVersionType, bibleCreator, false, &cardNo, &dbID);
	while (err != dmErrCantFind){
		tListP->numTranslations++;
		err = DmGetNextDatabaseByTypeCreator(false, &state, dmxVersionType, bibleCreator, false, &cardNo, &dbID);
	}
	if (err == dmErrCantFind) err = errNone;

	bswPrefsSize=4;
	bswPrefs=0;
	PrefGetAppPreferences('BSwb',0,&bswPrefs,&bswPrefsSize,true);
	if (bswPrefs & 1 == 1) {
		path = "/Palm/Programs/Bible+/";
		translationsCard = countcardTranslations(path);
		if (translationsCard == 0) {
			path = "/Palm/Programs/PPBL/";
			translationsCard = countcardTranslations(path);
			if (translationsCard == 0) {
				path = "/Palm/Launcher/";
				translationsCard = countcardTranslations(path);
			}
		}
	}

	if ((tListP->numTranslations + translationsCard) > 0) {
		mpTranslation = (bswTranslation*) MemPtrNew(sizeof(bswTranslation) * (tListP->numTranslations + translationsCard));
		mpTranslationNames = (Char**) MemPtrNew(sizeof(char*) * (tListP->numTranslations + 1 + translationsCard));
		mpTranslationCodes = (Char**) MemPtrNew(sizeof(char*) * (tListP->numTranslations + 1 + translationsCard));
		if ((mpTranslation!=0) && (mpTranslationNames!=0) && (mpTranslationCodes!=0)) {

			translationCount = 0;
			err = DmGetNextDatabaseByTypeCreator(true, &state, dmxVersionType, bibleCreator, false, &cardNo, &dbID);
			while ((err != dmErrCantFind) && translationCount < tListP->numTranslations){
				//open translation
				dRef = DmOpenDatabase(cardNo, dbID, dmModeReadOnly);
				if (dRef) {
					mhR0 = DmGetRecord(dRef,0);
					if (mhR0) {
						hdr = MemHandleLock(mhR0);
						if (hdr) {
							mpTranslationData = (TranslationData *) MemPtrNew(sizeof(TranslationData));
							MemPtrSetOwner(mpTranslationData,0);
							mpTranslation[translationCount].ownerData = mpTranslationData;			/* supplier of this translation can save any structure, eg to store references to dbID or databaseName, or VFS file name */
							if (mpTranslationData) {
								mpTranslationData->ordinal = translationCount+1;
								mpTranslationData->cardNo = cardNo;
								mpTranslationData->dbID = dbID;
								MemSet(mpTranslationData->fullName, 256, 0);
							}
							mpTranslation[translationCount].num = translationCount;							/* translation number */
							copyTranslationInfo(&(mpTranslation[translationCount]),(BibleHeaderRecord *)hdr);
							translationCount++;
							MemHandleUnlock(mhR0);
						}
						DmReleaseRecord(dRef,0,false);
					}
					DmCloseDatabase(dRef);
				}
				err = DmGetNextDatabaseByTypeCreator(false, &state, dmxVersionType, bibleCreator, false, &cardNo, &dbID);
			}
			if (err == dmErrCantFind) err = errNone;

			if (bswPrefs & 1 == 1) {
				if (!FtrGet(sysFileCVFSMgr, vfsFtrIDVersion, &dirIterator)){
					while ((volIterator != vfsIteratorStop) && translationCount < (tListP->numTranslations + translationsCard))  {
						err = VFSVolumeEnumerate(&volRefNum, &volIterator);
						if (err == errNone) {
							err = VFSFileOpen(volRefNum, path, vfsModeRead, &dirRef);
							if (err != errNone)
								continue;
							info.nameP = fileName; // point to local buffer
							info.nameBufLen = 256;
							dirIterator = vfsIteratorStart;
							while ((dirIterator != vfsIteratorStop) && translationCount < (tListP->numTranslations + translationsCard))  {
								// Get the next file
								err = VFSDirEntryEnumerate(dirRef, &dirIterator, &info);
								if (err != errNone)
									continue;
								MemSet(fullName, 256, 0);
								StrPrintF(fullName, "%s%s", path, fileName);
								if (VFSFileOpen(volRefNum, fullName, vfsModeRead, &fileRef))
									continue;
								if (VFSFileDBInfo(fileRef, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &fType, &fCreator, &totalRecord) || fType != dmxVersionType || fCreator != bibleCreator || !totalRecord) {
									VFSFileClose(fileRef);
									continue;
								}
								err = VFSFileDBGetRecord(fileRef,0,&mhR0,NULL,NULL);
								if (err != errNone)
									continue;
								if (mhR0) {
									hdr = MemHandleLock(mhR0);
									if (hdr) {
										mpTranslationData = (TranslationData *) MemPtrNew(sizeof(TranslationData));
										MemPtrSetOwner(mpTranslationData,0);
										mpTranslation[translationCount].ownerData = mpTranslationData;			/* supplier of this translation can save any structure, eg to store references to dbID or databaseName, or VFS file name */
										if (mpTranslationData) {
											mpTranslationData->ordinal = translationCount+1;
											mpTranslationData->cardNo = 0;
											mpTranslationData->dbID = 0;
											MemMove(mpTranslationData->fullName, fullName, 256);
										}
										mpTranslation[translationCount].num = translationCount;							/* translation number */
										copyTranslationInfo(&(mpTranslation[translationCount]),(BibleHeaderRecord *)hdr);
										translationCount++;
										MemHandleUnlock(mhR0);
									}
									MemHandleFree(mhR0);
								}
								VFSFileClose(fileRef);
							}
							VFSFileClose(dirRef);
						}
					}
				}
			}

			//some databases may have not opened correctly.
			tListP->numTranslations = translationCount;

			MemPtrSetOwner(mpTranslation,0);
			MemPtrSetOwner(mpTranslationNames,0);
			MemPtrSetOwner(mpTranslationCodes,0);
			tListP->translationArray = mpTranslation;
			i = 0;
			while (i < tListP->numTranslations) {
				mpTranslationNames[i] = mpTranslation[i].name;
				mpTranslationCodes[i] = mpTranslation[i].code;
				i++;
			}
			mpTranslationNames[i] = NULL;
			mpTranslationCodes[i] = NULL;

			tListP->forAPI = bswAPIMaskLaunch + bswAPIMaskGetReference + bswAPIMaskSelectedReference + bswAPIMaskGetCount;
			tListP->translationArray = mpTranslation;
			tListP->translationNames = mpTranslationNames;
			tListP->translationCodes = mpTranslationCodes;

		} else {
			if (mpTranslation) MemPtrFree(mpTranslation);
			if (mpTranslationNames) MemPtrFree(mpTranslationNames);
			if (mpTranslationCodes) MemPtrFree(mpTranslationCodes);
		}
	}
	
	return err;
}

static Err FreeTranslationList(bswTranslationList *tListP) {
	Err err=errNone;
	bswTranslation * mpTranslation;
	UInt16 i;
	UInt16 j;

	if ( tListP == NULL ) return bswErrParam;
	if ( tListP->structureSize != bswTranslationListStructureSizeV1 ) return bswErrBadStructureSize;

	if (tListP->translationArray) {
		for (i = 0; i < tListP->numTranslations; i++) {
			mpTranslation = &(tListP->translationArray[i]);
			if (mpTranslation) {
				if (mpTranslation->code) {
					MemPtrFree(mpTranslation->code);
					mpTranslation->code = NULL;
				}
				if (mpTranslation->name) {
					MemPtrFree(mpTranslation->name);
					mpTranslation->name = NULL;
				}
				if (mpTranslation->bookSeq) {
					MemPtrFree(mpTranslation->bookSeq);
					mpTranslation->bookSeq = NULL;
				}
				if (mpTranslation->booksChapters) {
					MemPtrFree(mpTranslation->booksChapters);
					mpTranslation->booksChapters = NULL;
				}
				if (mpTranslation->chaptersVerses) {
					for (j = 0; j < mpTranslation->numBooks; j++) {
						if (mpTranslation->chaptersVerses[j]) {
							MemPtrFree(mpTranslation->chaptersVerses[j]);
							mpTranslation->chaptersVerses[j] = NULL;
						}
					}
					MemPtrFree(mpTranslation->chaptersVerses);
					mpTranslation->chaptersVerses = NULL;
				}
				if (mpTranslation->ownerData) {
					// When this starts being used, more cleanup wil be required.
					MemPtrFree(mpTranslation->ownerData);
					mpTranslation->ownerData = NULL;
				}
			}
		}
		MemPtrFree(tListP->translationArray);
		tListP->translationArray = NULL;
	}

	tListP->numTranslations = 0;
	tListP->translationNames = NULL;
	tListP->translationCodes = NULL;

	return err;
}
