/* ChooseVerse.c */

#include <PalmOS.h>
#include <PalmOSGlue.h>
#include "bsw.h"
#include "BibleSwitch.h"

#include "ChooseVerse_Rsc.h"
#include "ChooseVerse.h"

#define noListSelection	-1
#define maxStrIToALen	12

UInt16 FldGetInsPtPosition(const FieldType *fldP) SYS_TRAP(sysTrapFldGetInsPtPosition);
Boolean FldInsert(FieldType *fldP,const Char *insertChars, UInt16 insertLen) SYS_TRAP(sysTrapFldInsert);
void FldSetInsPtPosition(FieldType *fldP, UInt16 pos) SYS_TRAP(sysTrapFldSetInsPtPosition);
void InsPtEnable(Boolean enableIt) SYS_TRAP(sysTrapInsPtEnable);

#define PROCESSEVENT 0x6000

#define STATUS_BOOK 1
#define STATUS_CHAPTER 2
#define STATUS_VERSE 3
#define STATUS_BUTTON 4

/* Initialize global variables */
static _global *InitGlobals()
{
    _global *g;
    g = MemPtrNew(sizeof(_global));
    return (g);
}

void LaunchBible(_global *g){
	bswReference ref;
	Err err;

	MemSet(&ref,sizeof(bswReference),0);
	ref.structureSize = sizeof(bswReference);

	ref.chapterNumber=g->chapterNum;
	ref.verseNumber=g->verseNum;
	err = bswGetBookID(g->refNum, g->clientContext, g->bookNum, &ref.bookNumber);
	if (errNone == err) {
		err = bswLaunch(g->refNum, &ref);
	}
}

void GetCurrentBibleRef(_global *g, bswReference * refP){
	bswReference ref;
	Err err;

	if ((refP == NULL) || (refP->bookNumber == 0)) {
		MemSet(&ref,sizeof(bswReference),0);
		ref.structureSize = sizeof(bswReference);
		err = bswSelectedReference(g->refNum, &ref);
		refP = &ref;
	}
	g->chapterNum = refP->chapterNumber;
	g->verseNum=refP->verseNumber;
	err = bswGetBooksSeqNum(g->refNum, g->clientContext, refP->bookNumber, &(g->bookNum));
}
static void setField(FieldType* fld, UInt16 Num)
{
	char sNumber[maxStrIToALen];
	StrIToA(sNumber,Num);
	FldSetSelection(fld,0,FldGetTextLength(fld));
	FldInsert(fld,sNumber,StrLen(sNumber));
}
static void setBook(_global *g, FormPtr frmP)
{
	UInt16 l;
	for(l=0;l<6;l++){
		LstSetSelection(FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP,l)),((g->bookNum>=g->FirstListBook[l]+1)&&(g->bookNum<=g->FirstListBook[l+1]))? g->bookNum-g->FirstListBook[l]-1:noListSelection);
	}
}
static void setScrollBars(_global *g){
	UInt16 bookID;
	UInt16 maxChapter;
	UInt16 maxVerse;
	Err err;
	err = bswGetBookID(g->refNum, g->clientContext, g->bookNum, &bookID);
	err = bswGetBooksNumChapters(g->refNum, g->clientContext, bookID, &maxChapter);
	if (g->chapterNum > maxChapter) {
		g->chapterNum = maxChapter;
		setField(g->fldC,g->chapterNum);
	}
	SclSetScrollBar(g->sbC,g->chapterNum,1,maxChapter,1);
	err = bswGetChaptersNumVerses(g->refNum, g->clientContext, bookID, g->chapterNum, &maxVerse);
	SclSetScrollBar(g->sbV,g->verseNum,1,maxVerse,1);
	if (g->verseNum > maxVerse) {
		g->verseNum = maxVerse;
		setField(g->fldV,g->verseNum);
	}
}
static void UpdateForm(_global *g, FormPtr frmP){
	setBook(g, frmP);
	setField(g->fldC,g->chapterNum);
	setField(g->fldV,g->verseNum);
	setScrollBars(g);
	InsPtEnable((g->status==STATUS_CHAPTER)||(g->status==STATUS_VERSE));
	if(g->status==STATUS_CHAPTER) {
		FldSetSelection(g->fldC,0,FldGetTextLength(g->fldC));
		FrmSetFocus(frmP,FrmGetObjectIndex(frmP,ChpF));
	}
	if(g->status==STATUS_VERSE) {
		FldSetSelection(g->fldV,0,FldGetTextLength(g->fldV));
		FrmSetFocus(frmP,FrmGetObjectIndex(frmP,VrsF));
	}
	CtlSetValue(g->okB,(g->status==STATUS_BUTTON));
}
static void ReadFields(_global *g){
	if(FldGetTextLength(g->fldC)>0){
		g->chapterNum=StrAToI(FldGetTextPtr(g->fldC));
	} else {
		g->chapterNum=1;
	}
	if(FldGetTextLength(g->fldV)>0){
		g->verseNum=StrAToI(FldGetTextPtr(g->fldV));
	} else {
		g->verseNum=1;
	}
}
static Boolean ProcessKey(_global *g, FormPtr frmP, EventType* event){
	WChar wc;
	Boolean h=false;
	EventType myEvent;
	wc=event->data.keyDown.chr;
	switch (wc) {
	case 10:	//next line
	case 13:	//select
	case 32:	//space
		if(g->status==STATUS_BUTTON) {
			//Launch bible reader
			ReadFields(g);
			setScrollBars(g);
			LaunchBible(g);
		} else {
			g->status++;
		}
		h=true;
		break;
	case 11:	//up hardware
	case 30:	//up
	case 268:	//up line
		switch(g->status) {
		case STATUS_BOOK:
			if(g->bookNum>1){
				g->bookNum-=1;
				g->chapterNum=1;
				g->verseNum=1;
				g->status=STATUS_BOOK;
			}
			break;
		case STATUS_CHAPTER:
			if(g->chapterNum>1){
				g->chapterNum-=1;
				g->verseNum=1;
				g->status=STATUS_CHAPTER;
			}
			break;
		case STATUS_VERSE:
			if(g->verseNum>1){
				g->verseNum-=1;
				g->status=STATUS_VERSE;
			}
			break;
		case STATUS_BUTTON:
			g->status=STATUS_VERSE;
			break;
		}
		h=true;
		break;
	case 12:	//down hardware
	case 31:	//down
	case 259:	//down line
		switch(g->status) {
		case STATUS_BOOK:
			if(g->bookNum<66){
				g->bookNum+=1;
				g->chapterNum=1;
				g->verseNum=1;
				g->status=STATUS_BOOK;
			}
			break;
		case STATUS_CHAPTER:
			if(g->chapterNum<150){
				g->chapterNum+=1;
				g->verseNum=1;
				g->status=STATUS_CHAPTER;
			}
			break;
		case STATUS_VERSE:
			if(g->verseNum<150){
				g->verseNum+=1;
				g->status=STATUS_VERSE;
			}
			break;
		case STATUS_BUTTON:
			g->status=STATUS_BOOK;
			break;
		}
		h=true;
		break;
	default:
		myEvent.eType=PROCESSEVENT;
		EvtAddEventToQueue(&myEvent);
		break;
	}
	if(h) {
		UpdateForm(g, frmP);
	}
	return h;
}
static void mainFormInit(_global *g, FormPtr frmP, bswReference * refP)
{
	g->fldC = FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP,ChpF));
	g->fldV = FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP,VrsF));

	g->sbC=FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP, ChpSB));
	g->sbV=FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP, VrsSB));

	g->okB=FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP, OKButton));
	
	g->bookNum=1;
	g->chapterNum=1;
	g->verseNum=1;
	
	GetCurrentBibleRef(g, refP);

	LstSetListChoices(FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP,0)), g->bookNames, 13);
	LstSetListChoices(FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP,1)), &(g->bookNames[13]), 13);
	LstSetListChoices(FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP,2)), &(g->bookNames[26]), 13);
	LstSetListChoices(FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP,3)), &(g->bookNames[39]), 9);
	LstSetListChoices(FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP,4)), &(g->bookNames[48]), 9);
	LstSetListChoices(FrmGetObjectPtr(frmP,FrmGetObjectIndex(frmP,5)), &(g->bookNames[57]), 9);
	
	g->status=STATUS_BOOK;
	
	FrmDrawForm(frmP);
	
	UpdateForm(g, frmP);
}

/*
 * FUNCTION: MainFormHandleEvent
 *
 * DESCRIPTION:
 *
 * This routine is the event handler for the "MainForm" of this 
 * application.
 *
 * PARAMETERS:
 *
 * eventP
 *     a pointer to an EventType structure
 *
 * RETURNED:
 *     true if the event was handled and should not be passed to
 *     FrmHandleEvent
 */

static Boolean MainFormHandleEvent(EventType * eventP)
{
	Boolean handled = false;
	UInt16				cardNo;
	LocalID				dbID;
	DmSearchStateType	searchState;
	UInt16 ctlNum;
	FormPtr frmP= FrmGetActiveForm();
	_global *g = FrmGetGadgetData(frmP, FrmGetObjectIndex(frmP, globalGadget));
	switch (eventP->eType) {
		case frmOpenEvent:
			mainFormInit(g, frmP, NULL);
			handled = true;
			break;
		case lstEnterEvent: 
			g->oldBookNum=g->bookNum;
			g->bookNum=-1;
			setBook(g, frmP);
			break;
		case lstExitEvent: 
			g->bookNum=g->oldBookNum;
			setBook(g, frmP);
			break;
		case lstSelectEvent: 
			ctlNum=eventP->data.lstSelect.listID;
			g->bookNum=1+eventP->data.lstSelect.selection + g->FirstListBook[ctlNum];
			g->chapterNum=1;
			g->verseNum=1;
			g->status=STATUS_CHAPTER;
			UpdateForm(g, frmP);
			handled=true;
			break;
		case fldEnterEvent:
			ctlNum=eventP->data.fldEnter.fieldID;
			switch(ctlNum){
				case ChpF:
					g->status=STATUS_CHAPTER;
					break;
				case VrsF:
					g->status=STATUS_VERSE;
					break;
			}
			CtlSetValue(g->okB,false);
			break;
		case ctlSelectEvent: 
			ctlNum=eventP->data.ctlSelect.controlID;
			switch(ctlNum){
				case Chp1:
					g->chapterNum=1;
					g->verseNum=1;
				case ChpL:
					g->status=STATUS_CHAPTER;
					UpdateForm(g, frmP);
					break;
				case Vrs1:
					g->verseNum=1;
				case VrsL:
					g->status=STATUS_VERSE;
					UpdateForm(g, frmP);
					break;
				case OKButton:
					//Launch bible reader
					ReadFields(g);
					setScrollBars(g);
					LaunchBible(g);
					break;
				case CancelButton:
					break;
			}
			handled=true;
			break;
		case sclRepeatEvent:
			switch(eventP->data.sclRepeat.scrollBarID){
			case ChpSB:
				g->chapterNum=eventP->data.sclRepeat.newValue;
				g->verseNum=1;
				g->status=STATUS_CHAPTER;
				break;
			case VrsSB:
				g->verseNum=eventP->data.sclRepeat.newValue;
				g->status=STATUS_VERSE;
				break;
			}
			UpdateForm(g, frmP);
			break;
		case keyDownEvent: 
			handled = ProcessKey(g, frmP, eventP);
			break;
		case PROCESSEVENT: 
			ReadFields(g);
			setScrollBars(g);
			handled=true;
			break;
		case menuEvent:
			switch (eventP->data.menu.itemID) {
				case HelpMenuPrefs:
					DmGetNextDatabaseByTypeCreator(true, &searchState, sysFileTPanel, prefsCreatorID, true, &cardNo, &dbID);
					if (dbID) {
						SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdPanelCalledFromApp, NULL);
					}
					handled = true;
					break;
				case HelpMenuAbout:
					MenuEraseStatus(NULL);
					FrmHelp(1000);
					handled = true;
					break;
			}
			break;
	}
	return handled;
}

/*
 * FUNCTION: AppHandleEvent
 *
 * DESCRIPTION: 
 *
 * This routine loads form resources and set the event handler for
 * the form loaded.
 *
 * PARAMETERS:
 *
 * event
 *     a pointer to an EventType structure
 *
 * RETURNED:
 *     true if the event was handled and should not be passed
 *     to a higher level handler.
 */

static Boolean AppHandleEvent(_global *g, EventType * eventP)
{
	UInt16 formId;
	FormType * frmP;
	

	if (eventP->eType == frmLoadEvent)
	{
		/* Load the form resource. */
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);

		/* 
		 * Set the event handler for the form.  The handler of the
		 * currently active form is called by FrmHandleEvent each
		 * time is receives an event. 
		 */
		switch (formId)
		{
			case MainForm:
				FrmSetGadgetData(frmP, FrmGetObjectIndex(frmP, globalGadget), g);
				FrmSetEventHandler(frmP, MainFormHandleEvent);
				break;

		}
		return true;
	}

	return false;
}

/* Event Handler */
static Boolean DAHandleEvent(EventPtr eventP, _global *g)
{
    Boolean done = false;
    Boolean handled = false;
    UInt16 ch;
	UInt16 ctlNum;
	UInt16				cardNo;
	LocalID				dbID;
	DmSearchStateType	searchState;
	FormPtr frmP = FrmGetActiveForm();

    switch (eventP->eType) {
	    case appStopEvent:
			/* Send appStopEvent to original application */
			EvtAddEventToQueue(eventP);
			g->cancelled = true;
			done = true;
			handled = true;
			break;
	    case keyDownEvent:
			ch = eventP->data.keyDown.chr;
			if (TxtCharIsHardKey(eventP->data.keyDown.modifiers, eventP->data.keyDown.chr)) {
			    /* maybe 'HotSync!' button pressed ? */
			    EvtAddEventToQueue(eventP);
			    g->cancelled = true;
			    done = true;
			    handled = true;
			} else {
				handled = ProcessKey(g, frmP, eventP);
			}
			break;
		case lstEnterEvent: 
			g->oldBookNum=g->bookNum;
			g->bookNum=-1;
			setBook(g, frmP);
			break;
		case lstExitEvent: 
			g->bookNum=g->oldBookNum;
			setBook(g, frmP);
			break;
		case lstSelectEvent: 
			ctlNum=eventP->data.lstSelect.listID;
			g->bookNum=1+eventP->data.lstSelect.selection + g->FirstListBook[ctlNum];
			g->chapterNum=1;
			g->verseNum=1;
			g->status=STATUS_CHAPTER;
			UpdateForm(g, frmP);
			handled=true;
			break;
		case fldEnterEvent:
			ctlNum=eventP->data.fldEnter.fieldID;
			switch(ctlNum){
				case ChpF:
					g->status=STATUS_CHAPTER;
					break;
				case VrsF:
					g->status=STATUS_VERSE;
					break;
			}
			CtlSetValue(g->okB,false);
			break;
		case ctlSelectEvent: 
			ctlNum=eventP->data.ctlSelect.controlID;
			switch(ctlNum){
				case Chp1:
					g->chapterNum=1;
					g->verseNum=1;
				case ChpL:
					g->status=STATUS_CHAPTER;
					UpdateForm(g, frmP);
					break;
				case Vrs1:
					g->verseNum=1;
				case VrsL:
					g->status=STATUS_VERSE;
					UpdateForm(g, frmP);
					break;
				case OKButton:
					//Launch bible reader
					ReadFields(g);
					setScrollBars(g);
					g->cancelled = false;
				    done = true;
					break;
				case CancelButton:
					g->cancelled = true;
				    done = true;
					break;
			}
		    handled = true;
			break;
		case sclRepeatEvent:
			switch(eventP->data.sclRepeat.scrollBarID){
			case ChpSB:
				g->chapterNum=eventP->data.sclRepeat.newValue;
				g->verseNum=1;
				g->status=STATUS_CHAPTER;
				break;
			case VrsSB:
				g->verseNum=eventP->data.sclRepeat.newValue;
				g->status=STATUS_VERSE;
				break;
			}
			UpdateForm(g, frmP);
			break;
		case PROCESSEVENT: 
			ReadFields(g);
			setScrollBars(g);
			handled=true;
			break;
		case menuEvent:
			switch (eventP->data.menu.itemID) {
				case HelpMenuPrefs:
					DmGetNextDatabaseByTypeCreator(true, &searchState, sysFileTPanel, prefsCreatorID, true, &cardNo, &dbID);
					if (dbID) {
						SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdPanelCalledFromApp, NULL);
					}
					handled = true;
					break;
				case HelpMenuAbout:
					MenuEraseStatus(NULL);
					FrmHelp(1000);
					handled = true;
					break;
			}
			break;
	    default:
			break;
    }
    if (!handled) {
		FrmHandleEvent(FrmGetActiveForm(), eventP);
    }
    return (done);
}

/* Event Loop */
static void EventLoop(_global *g)
{
    EventType event;
    Boolean done = false;
    UInt16 error;
    do {
		EvtGetEvent(&event, evtWaitForever);
		if (SysHandleEvent(&event))
		    continue;
		if (MenuHandleEvent(NULL, &event, &error))
		    continue;
		done = DAHandleEvent(&event, g);
    } while (!done);
}

/*
 * FUNCTION: AppEventLoop
 *
 * DESCRIPTION: This routine is the event loop for the application.
 */

static void AppEventLoop(_global *g)
{
	UInt16 error;
	EventType event;

	do 
	{
		/* change timeout if you need periodic nilEvents */
		EvtGetEvent(&event, evtWaitForever);

		if (! SysHandleEvent(&event))
		{
			if (! MenuHandleEvent(0, &event, &error))
			{
				if (! AppHandleEvent(g, &event))
				{
					FrmDispatchEvent(&event);
				}
			}
		}
	} while (event.eType != appStopEvent);
}

/*
 * FUNCTION: AppStart
 *
 * DESCRIPTION:  Get the current application's preferences.
 *
 * RETURNED:
 *     errNone - if nothing went wrong
 */

static Err AppStart(_global *g)
{
	Err err;
	g->FirstListBook[0] = 0;
	g->FirstListBook[1] = 13;
	g->FirstListBook[2] = 26;
	g->FirstListBook[3] = 39;
	g->FirstListBook[4] = 48;
	g->FirstListBook[5] = 57;
	g->FirstListBook[6] = 66;

	err = bsw_OpenLibrary(&(g->refNum), &(g->clientContext));
	if (errNone == err) {
		err = bswGetBookNames(g->refNum, g->clientContext, bswShortTextReferenceType, &(g->bookNames), &(g->namesResource));
	}
	return err;
}

/*
 * FUNCTION: AppStop
 *
 * DESCRIPTION: Save the current state of the application.
 */

static void AppStop(_global *g)
{
	if (g->namesResource) bswFreeBookNames(g->refNum, g->clientContext, &(g->bookNames), &(g->namesResource));
	bsw_CloseLibrary(g->refNum, g->clientContext);
}

Err SelectReference(bswReference * refP) {
    _global *g;
	Err err = errNone;
    FormPtr form, oldForm;

	if (refP == NULL) return bswErrParam;
	if( refP->structureSize != bswReferenceStructureSizeV1 ) return bswErrBadStructureSize;

    /* Allocate && initialize global variables */
    g = InitGlobals();
    if (g == NULL) {
		/* Fail to allocate global variables area */
		return bswErrLock;
    }

	err = AppStart(g);
	if (err == errNone) {
	    oldForm = FrmGetActiveForm();
	    /* Activate Main Form */
	    form  = FrmInitForm(MainForm);
		mainFormInit(g, form, refP);
	    FrmSetActiveForm(form);
	    FrmDrawForm(form);

	    /* goto event handler */
	    EventLoop(g);
	    
	    if (g->cancelled == false) {
			err = bswGetBookID(g->refNum, g->clientContext, g->bookNum, &(refP->bookNumber));
	    	refP->chapterNumber = g->chapterNum;
	    	refP->verseNumber = g->verseNum;
	    }

	    /* finish up application */
	    FrmEraseForm(form);
	    FrmDeleteForm(form);
	    FrmSetActiveForm(oldForm);

		AppStop(g);
	}

    /* Release Global area */
    MemPtrFree(g);

    return err;
}

Err NormalLaunch() {
    _global *g;
	Err err = errNone;

    /* Allocate && initialize global variables */
    g = InitGlobals();
    if (g == NULL) {
		/* Fail to allocate global variables area */
		return bswErrLock;
    }

	err = AppStart(g);
	if (err) 
		return err;

	/* 
	 * start application by opening the main form
	 * and then entering the main event loop 
	 */
	FrmGotoForm(MainForm);
	AppEventLoop(g);

	// Close all the open forms.
	FrmCloseAllForms();

	AppStop(g);

    /* Release Global area */
    MemPtrFree(g);

    return errNone;
}

UInt32 PilotMain(UInt16 launchCode, void * cmdPBP, UInt16 launchFlags){
	#pragma unused(launchFlags)
	Err err = errNone;
	switch (launchCode){
		case sysAppLaunchCmdNormalLaunch:
			err = NormalLaunch();
			break;
		case bswAppLaunchCmdSelectReference:
			err = SelectReference((bswReference *)cmdPBP);
			break;
	}
	return err;
} 
