/* ShowVerse.c */
#ifndef __MWERKS__
#include "ShowVerseOBC.h"
#endif
#ifdef __MWERKS__
#include <PalmOS.h>
#include <PalmOSGlue.h>
#endif
#include "bsw.h"
#include "ShowVerse_Rsc.h"
#include "ShowVerseBMP_Rsc.h"
#include "ShowVerse.h"

/* Define the minimum OS version we support */
#define ourMinVersion    sysMakeROMVersion(3,0,0,0)

/*********************************************************************
 * Internal Functions
 *********************************************************************/

#define MAXREFERENCELENGTH 32

static void MainFormInit(FormType *frmP)
{
	UInt16 currentPrefs;
	UInt16 prefsSize;

	prefsSize=2;
	currentPrefs=0;
	PrefGetAppPreferences(appFileCreator,1,&currentPrefs,&prefsSize,true);
	if (currentPrefs==0) {
		FrmSetControlGroupSelection(frmP, 1, offButton);
	} else {
		FrmSetControlGroupSelection(frmP, 1, onButton);
	}
}

// preference
typedef struct _prefs {
	Int16 RefType;
	Int16 RefPosition;
	Int16 NoRefAction;
	Int16 Translation;
	Int16 ShowTrans;
} _prefs;

// global variable
typedef struct _global {
	_prefs prefs;
	FormPtr form;
	FieldPtr field;
	bswReference ref;
	MemHandle textHandle;
	char* textRef;
	Boolean done;
	UInt32 lastRepeatEventTime;
	UInt16 libRefNum;
	UInt32 libClientContext;
} _global;

static void PrefsEventLoop(_global *g);

/*========================================================================*/

static void *GetObjectPtr(FormPtr form, UInt16 objectID)
{
    UInt16     index;
    void    *ptr;

    index = FrmGetObjectIndex(form, objectID);
    ptr   = FrmGetObjectPtr(form, index);
    return (ptr);
}


/* Initialize Preference */
static void InitPrefs(_global *g)
{
    MemSet(&g->prefs, sizeof(_prefs), 0);
}

/* Load Preference */
static void LoadPrefs(_global *g)
{
    Int16 version;
    UInt16 prefSize = sizeof(_prefs);

    version = PrefGetAppPreferences(appFileCreator, 2, &g->prefs, &prefSize, true);
    if (version == noPreferenceFound) {
		/* No preference found, create it with default value */
		InitPrefs(g);
    }
}

/* Save Preference */
static void SavePrefs(_global *g)
{
    PrefSetAppPreferences(appFileCreator, 2, 0, &g->prefs, sizeof(g->prefs), true);
}

/* Initialize global variables */
static _global *InitGlobals()
{
    _global *g;
    g = MemPtrNew(sizeof(_global));
    return (g);
}
static void UpdateScrollBar(FieldType *field, ScrollBarType *bar){
 	UInt16 spos, th, fh, mx;
 	FldGetScrollValues(field, &spos, &th, &fh);
	if(th>fh) {
		mx=th-fh;
	} else {
		if(spos){
			mx=spos;
		} else {
			mx=0;
		}
	}
	SclSetScrollBar(bar, spos, 0, mx, fh-1);
}
static char* GetSelection(FormPtr oldForm, _global *g) {
	UInt16 fcs = 0;
	FieldType *oldField;
	UInt16 fldType;
	UInt16 startPos, endPos;
	Int16 selLength;
	char *selection = NULL;
	char *p;
	fcs = FrmGetFocus(oldForm);
	if (fcs != noFocus) {
		oldField = FrmGetObjectPtr(oldForm,fcs);
		if (oldField != NULL) {
			fldType = FrmGetObjectType(oldForm, fcs);
			if (fldType == frmFieldObj) {
				FldGetSelection(oldField, &startPos, &endPos);
				selLength = endPos - startPos;
				if (selLength > 0) {
					selection = MemPtrNew(selLength + 1);
					if (selection) {
						p = FldGetTextPtr(oldField);
						StrNCopy(selection, &(p[startPos]), selLength);
						selection[selLength] = 0;
					}
				}
			}
		}
	}
	return selection;
}
static void SetField(_global *g){
	char *p;
	bswText text;
	Err err;
	if (g->textHandle != NULL) {
		FldSetTextHandle(g->field, NULL);
		MemHandleFree(g->textHandle);
	}
	g->textHandle = NULL;
	if (g->textRef != NULL) {
		MemPtrFree(g->textRef);
	}
	g->textRef = NULL;
	if (g->prefs.RefType == 0) {
		err = bswCompleteRef(g->libRefNum, g->libClientContext, &(g->ref), bswStandardReferenceType, bswStandardReferenceType + bswTextReferenceType);
	}
	if (g->prefs.RefType == 1) {
		err = bswCompleteRef(g->libRefNum, g->libClientContext, &(g->ref), bswStandardReferenceType, bswStandardReferenceType + bswShortTextReferenceType);
	}
	if (g->prefs.RefType == 2) {
		err = bswCompleteRef(g->libRefNum, g->libClientContext, &(g->ref), bswStandardReferenceType, bswStandardReferenceType + bswTinyTextReferenceType);
	}
	MemSet(&text,sizeof(bswText),0);
	text.structureSize = sizeof(bswText);
	text.refP = &(g->ref);
	err = bswGetReference(g->libRefNum, &text);
	g->textRef =g->ref.textRef;
	g->textHandle = MemHandleNew(1024);
	p = MemHandleLock(g->textHandle);
	MemSet(p,1024,0);
	if (g->textRef) {
		if (g->prefs.RefPosition == 0) {
			StrCat(p, g->textRef);
			StrCat(p, " ");
		}
		if (text.text) {
			StrCat(p, text.text);
		}
		if (g->prefs.RefPosition == 1) {
			StrCat(p, " (");
			StrCat(p, g->textRef);
			StrCat(p, ")");
		}
	}
	MemHandleUnlock(g->textHandle);
	if (text.text) {
		MemPtrFree(text.text);
	}
	FldSetTextHandle(g->field, (MemHandle)g->textHandle);
	UpdateScrollBar(g->field,GetObjectPtr(g->form, DAScrollBar));
}
static void ShowText(FormPtr oldForm, _global *g) {
	Err err;
	char *selection = NULL;
	char *p;

	g->ref.structureSize = sizeof(bswReference);

	selection = GetSelection(oldForm, g);
	g->ref.textRef = selection;
	if (selection) {
		err = bswCompleteRef(g->libRefNum, g->libClientContext, &(g->ref), bswTextReferenceType, bswStandardReferenceType);
		if (g->ref.bookNumber == 0) {
			g->textHandle = MemHandleNew(1024);
			p = MemHandleLock(g->textHandle);
			StrCopy(p, "Unknown Reference:");
			StrCat(p, "\n");
			StrCat(p, selection);
			MemHandleUnlock(g->textHandle);
			FldSetTextHandle(g->field, (MemHandle)g->textHandle);
			UpdateScrollBar(g->field,GetObjectPtr(g->form, DAScrollBar));
		} else {
			SetField(g);
		}
	 	MemPtrFree(selection);
	} else {
		switch (g->prefs.NoRefAction) {
		case 0:
			err = bswSelectedReference(g->libRefNum, &(g->ref));
			if (g->ref.bookNumber == 0) {
				g->textHandle = MemHandleNew(1024);
				p = MemHandleLock(g->textHandle);
				StrCopy(p, "No Selection");
				MemHandleUnlock(g->textHandle);
				FldSetTextHandle(g->field, (MemHandle)g->textHandle);
				UpdateScrollBar(g->field,GetObjectPtr(g->form, DAScrollBar));
			} else {
				SetField(g);
			}
			break;
		case 1:
			err = bswSelectReference(g->libRefNum, &(g->ref));
		 	SetField(g);
			break;
		default:
			g->textHandle = MemHandleNew(1024);
			p = MemHandleLock(g->textHandle);
			StrCopy(p, "No Selection");
			MemHandleUnlock(g->textHandle);
			FldSetTextHandle(g->field, (MemHandle)g->textHandle);
			UpdateScrollBar(g->field,GetObjectPtr(g->form, DAScrollBar));
		}
	}
}

/* Event Handler */
static Boolean DAHandlePrefsEvent(EventPtr event, _global *g)
{
    Boolean done = false;
    Boolean handled = false;
    FormPtr form;
    WinHandle winH;
    RectangleType r;
    UInt16 ch;
    switch (event->eType) {
	    case appStopEvent:
			/* Send appStopEvent to original application */
			EvtAddEventToQueue(event);
			done = true;
			g->done=true;
			handled = true;
			break;
	    case keyDownEvent:
			ch = event->data.keyDown.chr;
			if (TxtCharIsHardKey(event->data.keyDown.modifiers, event->data.keyDown.chr)) {
			    /* maybe 'HotSync!' button pressed ? */
			    EvtAddEventToQueue(event);
			    done = true;
			    g->done=true;
			    handled = true;
			}
			break;
	    case penDownEvent:
			form = FrmGetActiveForm();
			winH = FrmGetWindowHandle(form);
			FrmGetFormBounds(form,&r);
		    /* screenX and screenY are window relative */
			if (!RctPtInRectangle(event->screenX + r.topLeft.x, event->screenY + r.topLeft.y, &r)) {
				/* Pen tapped on Outside of form -- quit */
			    done = true;
			    g->done=true;
			    handled = true;
			}
			break;
		case ctlSelectEvent:
			switch (event->data.ctlSelect.controlID){
				case DAPrefsRefFull:
					g->prefs.RefType = 0;
					break;
				case DAPrefsRefShort:
					g->prefs.RefType = 1;
					break;
				case DAPrefsRefTiny:
					g->prefs.RefType = 2;
					break;
				case DAPrefsPosBefore:
					g->prefs.RefPosition = 0;
					break;
				case DAPrefsPosAfter:
					g->prefs.RefPosition = 1;
					break;
				case DAPrefsPosNotShown:
					g->prefs.RefPosition = 2;
					break;
				case DAPrefsNoRefPrev:
					g->prefs.NoRefAction = 0;
					break;
				case DAPrefsNoRefSelect:
					g->prefs.NoRefAction = 1;
					break;
				case DAPrefsNoRefHide:
					g->prefs.NoRefAction = 2;
					break;
				case DAPrefsShowTrans:
					g->prefs.ShowTrans = event->data.ctlSelect.on;
					break;
				case DAPrefsOKButton:
			  		done = true;
					handled = true;
					break;
			}
			break;
	    default:
			break;
    }
    if (!handled) {
		FrmHandleEvent(FrmGetActiveForm(), event);
    }
    return (done);
}
static void PrevVerse(_global *g){
	Err err;
	UInt16 seqNum;
	if (g->ref.bookNumber > 0) {
		if (g->ref.verseNumber > 1) {
			g->ref.verseNumber--;
		} else {
			if (g->ref.chapterNumber > 1) {
				g->ref.chapterNumber--;
			} else {
				err = bswGetBooksSeqNum(g->libRefNum, g->libClientContext, g->ref.bookNumber, &seqNum);
				if (seqNum > 1) {
					seqNum--;
				} else {
					err = bswGetTotalNumBooks(g->libRefNum, g->libClientContext, &seqNum);
				}
				err = bswGetBookID(g->libRefNum, g->libClientContext, seqNum, &g->ref.bookNumber);
				g->ref.chapterNumber = -1;
			}
			g->ref.verseNumber = -1;
		}
	 	SetField(g);
	 	FldDrawField(g->field);
	}
}
static void NextVerse(_global *g){
	Err err;
	UInt16 max;
	UInt16 seqNum;
	if (g->ref.bookNumber > 0) {
		err = bswGetChaptersNumVerses(g->libRefNum, g->libClientContext, g->ref.bookNumber, g->ref.chapterNumber, &max);
		if ( g->ref.verseNumber >= max) {
			err = bswGetBooksNumChapters(g->libRefNum, g->libClientContext, g->ref.bookNumber, &max);
			if ( g->ref.chapterNumber >= max) {
				err = bswGetBooksSeqNum(g->libRefNum, g->libClientContext, g->ref.bookNumber, &seqNum);
				err = bswGetTotalNumBooks(g->libRefNum, g->libClientContext, &max);
				if ( seqNum >= max) {
					seqNum = 1;
				} else {
					seqNum++;
				}
				err = bswGetBookID(g->libRefNum, g->libClientContext, seqNum, &g->ref.bookNumber);
				g->ref.chapterNumber = 1;
			} else {
				g->ref.chapterNumber++;
			}
			g->ref.verseNumber = 1;
		} else {
			g->ref.verseNumber++;
		}
	 	SetField(g);
		FldDrawField(g->field);
	}
}
static void ChooseVerse(_global *g){
	Err err;
	err = bswSelectReference(g->libRefNum, &(g->ref));
 	SetField(g);
	FldDrawField(g->field);
}
static void DPrefs(_global *g){
	Err err;
	FormPtr form, oldForm;
	bswCount countOfTranslations;
	/* Activate Main Form */
	oldForm = FrmGetActiveForm();
	form  = FrmInitForm(IdDAPrefsForm);
	FrmSetActiveForm(form);
	switch(g->prefs.RefType) {
		case 1:
			FrmSetControlGroupSelection(form, 1, DAPrefsRefShort);
			break;
		case 2:
			FrmSetControlGroupSelection(form, 1, DAPrefsRefTiny);
			break;
		default:
			FrmSetControlGroupSelection(form, 1, DAPrefsRefFull);
			break;
	}
	switch(g->prefs.RefPosition) {
		case 1:
			FrmSetControlGroupSelection(form, 2, DAPrefsPosAfter);
			break;
		case 2:
			FrmSetControlGroupSelection(form, 2, DAPrefsPosNotShown);
			break;
		default:
			FrmSetControlGroupSelection(form, 2, DAPrefsPosBefore);
			break;
	}
	switch(g->prefs.NoRefAction) {
		case 1:
			FrmSetControlGroupSelection(form, 3, DAPrefsNoRefSelect);
			break;
		case 2:
			FrmSetControlGroupSelection(form, 3, DAPrefsNoRefHide);
			break;
		default:
			FrmSetControlGroupSelection(form, 3, DAPrefsNoRefPrev);
			break;
	}

	countOfTranslations.structureSize=bswCountStructureSizeV1;
	countOfTranslations.flagWhatToCount=bswContextTranslation;
	countOfTranslations.withinContextReference=NULL;
	countOfTranslations.flagWithinWhatContext=bswContextEmpty;
	countOfTranslations.countResult=-1;
	err = bswGetCount(g->libRefNum, g->libClientContext, &countOfTranslations);

	FrmDrawForm(form);

	/* goto event handler */
	PrefsEventLoop(g);
	FrmEraseForm(form);
	FrmDeleteForm(form);
	FrmSetActiveForm(oldForm);
}
static void Launch(_global *g){
	Err err;
	if (g->ref.bookNumber > 0) {
		err = bswLaunch(g->libRefNum, &(g->ref));
	}
}
static Boolean DAHandleEvent(EventPtr event, _global *g) {
    Boolean done = false;
    Boolean handled = false;
    FormPtr form;
    WinHandle winH;
    //RectangleType r;
    FieldPtr field;
    UInt16 ch;
 	UInt16 spos, th, fh;
	UInt16 startPos, endPos;

    switch (event->eType) {
	    case appStopEvent:
			/* Send appStopEvent to original application */
			EvtAddEventToQueue(event);
			done = true;
			handled = true;
			break;
	    case winExitEvent:
			/* Pen tapped on Outside of form -- quit */
			winH = FrmGetWindowHandle(g->form);
			if (winH==event->data.winExit.exitWindow) {
			    done = true;
			    handled = true;
			}
			break;
	    case keyDownEvent:
			ch = event->data.keyDown.chr;
			if (TxtCharIsHardKey(event->data.keyDown.modifiers, event->data.keyDown.chr)) {
			    /* maybe 'HotSync!' button pressed ? */
			    EvtAddEventToQueue(event);
			    done = true;
			    handled = true;
			} else {
				switch (event->data.keyDown.chr) {
					case chrEscape:	//cancel
					    done = true;
					    handled = true;
					    break;
					case chrLeftArrow:
						PrevVerse(g);
						handled=true;
						break;
					case vchrPageUp:
					case chrUpArrow:
					case vchrPrevField:
						form = FrmGetActiveForm();
						field = GetObjectPtr(form, IdVerseField);
						FldGetScrollValues(field, &spos, &th, &fh);
						if (spos>0) {
							FldScrollField(field,fh-1,winUp);
							UpdateScrollBar(field,GetObjectPtr(form, DAScrollBar));
						} else {
							PrevVerse(g);
						}
						handled=true;
						break;
					case chrRightArrow:
						NextVerse(g);
						handled=true;
						break;
					case vchrPageDown:
					case chrDownArrow:
					case vchrNextField:
						form = FrmGetActiveForm();
						field = GetObjectPtr(form, IdVerseField);
						FldGetScrollValues(field, &spos, &th, &fh);
						if (th>(spos + fh)) {
							FldScrollField(field,fh-1,winDown);
							UpdateScrollBar(field,GetObjectPtr(form, DAScrollBar));
						} else {
							NextVerse(g);
						}
						handled=true;
						break;
					case 'p':
						DPrefs(g);
					 	SetField(g);
						FldDrawField(g->field);
						done=g->done;
				   		handled = true;
						break;
				}
			}
			break;
//	    case penDownEvent:
//			form = FrmGetActiveForm();
//			winH = FrmGetWindowHandle(form);
//			FrmGetFormBounds(form,&r);
//		    /* screenX and screenY are window relative */
//			if (!RctPtInRectangle(event->screenX + r.topLeft.x, event->screenY + r.topLeft.y, &r)) {
//				/* Pen tapped on Outside of form -- quit */
//			    done = true;
//			    handled = true;
//			}
//			break;
/*		case ctlEnterEvent:
			switch (event->data.ctlEnter.controlID){
				case DAPrevButton:
				case DANextButton:
					g->lastRepeatEventTime = 0;
					break;
			}
			break;
*/		case ctlRepeatEvent:
			//if (event->data.ctlRepeat.time - g->lastRepeatEventTime >= (SysTicksPerSecond() / 2)) {
				switch (event->data.ctlRepeat.controlID){
					case DAPrevButton:
						PrevVerse(g);
						break;
					case DANextButton:
						NextVerse(g);
						break;
				}
			//	g->lastRepeatEventTime = event->data.ctlRepeat.time;
			//}
			break;
		case ctlSelectEvent:
			switch (event->data.ctlSelect.controlID){
				case DACopyButton:
					form = FrmGetActiveForm();
					field = GetObjectPtr(form, IdVerseField);
					FldGetSelection(field, &startPos, &endPos);
					if (endPos - startPos == 0) {
						FldSetSelection(field, 0, FldGetTextLength(field));
					}
					FldCopy(field);
				   	handled = true;
					break;
				case DAChooseButton:
					ChooseVerse(g);
			   		handled = true;
					break;
				case DALaunchButton:
					Launch(g);
					done = true;
					handled = true;
					break;
			}
			break;
		case fldChangedEvent:
			form = FrmGetActiveForm();
			field = GetObjectPtr(form, IdVerseField);
			UpdateScrollBar(field,GetObjectPtr(form, DAScrollBar));
			handled = true;
			break;
		case sclRepeatEvent:
			form = FrmGetActiveForm();
			field = GetObjectPtr(form, IdVerseField);
			FldGetScrollValues(field, &spos, &th, &fh);
			if(spos>event->data.sclRepeat.newValue) {
				FldScrollField(field,spos-event->data.sclRepeat.newValue,winUp);
			} else {
				if(spos<event->data.sclRepeat.newValue) {
					FldScrollField(field,event->data.sclRepeat.newValue-spos,winDown);
				}
			}
			break;
		default:
			break;
	}
	if (!handled) {
		FrmHandleEvent(FrmGetActiveForm(), event);
	}
	return (done);
}
/* Event Loops */
static void EventLoop(_global *g) {
	EventType event;
	Boolean done = false;
	UInt16 error;
	do {
		EvtGetEvent(&event, evtWaitForever);
		if (SysHandleEvent(&event))
			continue;
		if (MenuHandleEvent(NULL, &event, &error))
			continue;
		done = DAHandleEvent(&event, g);
	} while (!done);
}
static void PrefsEventLoop(_global *g) {
	EventType event;
	Boolean done = false;
	UInt16 error;
	do {
		EvtGetEvent(&event, evtWaitForever);
		if (SysHandleEvent(&event))
			continue;
		if (MenuHandleEvent(NULL, &event, &error))
			continue;
		done = DAHandlePrefsEvent(&event, g);
	} while (!done);
}
void PerformDA() {
	_global *g;
	Err err;
	FormPtr oldForm;
	UInt32 feature = 0;

	/* Check duplicate invocation of FindVerse */
	if (FtrGet(appFileCreator, 0, &feature) == 0) {
		/* Already FindVerse is invoked -- duplicate invocation */
		return;
	} else {
		/* To avoid duplicate invocation, set feature */
		FtrSet(appFileCreator, 0, feature);
	}
	/* Allocate && initialize global variables */
	g = InitGlobals();
	if (g == NULL) {
		/* Fail to allocate global variables area */
		FtrUnregister(appFileCreator, 0);
		return;
	}
	MemSet(g, sizeof(_global), 0);
	/* Load Preferences from last invoked */
	err = bsw_OpenLibrary(&(g->libRefNum), &(g->libClientContext));
	if (err == errNone) {
		LoadPrefs(g);
		oldForm = FrmGetActiveForm();
		/* Activate Main Form */
		g->form  = FrmInitForm(IdDAForm);
		FrmSetActiveForm(g->form);
		/* Show verse text */
		g->field = GetObjectPtr(g->form, IdVerseField);
		MenuEraseStatus(NULL);
		ShowText(oldForm, g);
		FrmDrawForm(g->form);
		/* goto event handler */
		EventLoop(g);
		/* finish up application */
		FldSetTextHandle(g->field, NULL);
		if (g->textHandle != NULL) {
			MemHandleFree(g->textHandle);
		}
		g->textHandle = NULL;
		if (g->textRef != NULL) {
			MemPtrFree(g->textRef);
		}
		g->textRef = NULL;
		SavePrefs(g);
		FrmEraseForm(g->form);
		FrmDeleteForm(g->form);
		FrmSetActiveForm(oldForm);
		err = bsw_CloseLibrary(g->libRefNum, g->libClientContext);
	} /* TODO: else report problem? */
	/* Release feature */
	FtrUnregister(appFileCreator, 0);
	/* Release Global area */
	MemPtrFree(g);
}

/*
 *	Could cycle through DmGetNextOpenResDatabase, using DmOpenDatabaseInfo to get card & dbID,
 *	or set a feature & check whether it is still a dbRef pointer?
 *	when to turn it off?
 *	create recyclable db?
 *
 */

static void AddMenu(void) {
	UInt16 cardNo;
	LocalID dbID;
	DmOpenRef dbRef;
	Err err = errNone;
 	UInt16 currentPrefs;
	UInt16 prefsSize;

	prefsSize=2;
	currentPrefs=0;
	PrefGetAppPreferences(appFileCreator,1,&currentPrefs,&prefsSize,true);
	if (currentPrefs==1) {
		err = SysCurAppDatabase(&cardNo,&dbID);
		err = SysNotifyRegister(cardNo, dbID, sysNotifyMenuCmdBarOpenEvent, NULL, sysNotifyNormalPriority, NULL);
		err = SysNotifyRegister(cardNo, dbID, appFileCreator, NULL, sysNotifyNormalPriority, NULL);
		dbRef = DmOpenDatabaseByTypeCreator('Rsrc', appFileCreator, dmModeReadOnly + dmModeLeaveOpen);
	}
}
static void RemoveMenu(void) {
	UInt16 cardNo;
	LocalID dbID;
	Err err = errNone;

	err = SysCurAppDatabase(&cardNo,&dbID);
	err = SysNotifyUnregister(cardNo, dbID, sysNotifyMenuCmdBarOpenEvent, sysNotifyNormalPriority);
	err = SysNotifyUnregister(cardNo, dbID, appFileCreator, sysNotifyNormalPriority);
}
static void AddMyCommandBarItem() {
	Err err = errNone;
	UInt32 feature = 0;
	/* Check duplicate invocation of FindVerse */
	if (FtrGet(appFileCreator, 0, &feature) == 0) {
		return;
	} else {
		err = MenuCmdBarAddButton(menuCmdBarOnLeft, 9876, menuCmdBarResultNotify, appFileCreator, "Show Verse");
	}
}
static Boolean MainFormHandleEvent(EventType * eventP)
{
	Boolean handled = false;
	FormType * frmP;
 	FormPtr uForm;
	FieldPtr testField;

	UInt16 currentPrefs;

	UInt16				cardNo;
	LocalID				dbID;
	DmSearchStateType	searchState;

	Err err = errNone;

	switch (eventP->eType)
	{
		case frmOpenEvent:
			frmP = FrmGetActiveForm();
			MainFormInit(frmP);
			FrmDrawForm(frmP);
			handled = true;
			break;
		case ctlSelectEvent:
			switch (eventP->data.ctlSelect.controlID){
				case aboutButton:
					FrmAlert(AboutAlert);
					handled = true;
					break;
				case usageButton:
					uForm = FrmInitForm(usageForm);
					testField = GetObjectPtr(uForm, TestField);
					FldInsert(testField,"Try Jn 3:16 here",16);
					//FldSetInsPtPosition(testField, 11);
					FldSetSelection(testField, 4, 11);
					FrmSetFocus(uForm,FrmGetObjectIndex(uForm,TestField));
					FrmDoDialog(uForm);
					FrmDeleteForm(uForm);
					handled = true;
					break;
				case onButton:
					currentPrefs = 1;
					PrefSetAppPreferences(appFileCreator,1,1,&currentPrefs,2,true);
					if (currentPrefs) {
						AddMenu();
					} else {
						RemoveMenu();
					}
					handled = true;
					break;
				case offButton:
					currentPrefs = 0;
					PrefSetAppPreferences(appFileCreator,1,1,&currentPrefs,2,true);
					if (currentPrefs) {
						AddMenu();
					} else {
						RemoveMenu();
					}
					handled = true;
					break;
				case prefsButton:
					DmGetNextDatabaseByTypeCreator(true, &searchState, sysFileTPanel, prefsCreatorID, true, &cardNo, &dbID);
					if (dbID) {
						SysUIAppSwitch(cardNo, dbID, sysAppLaunchCmdPanelCalledFromApp, NULL);
					}
					handled = true;
					break;
			}
			break;
	}
	return handled;
}
static Boolean AppHandleEvent(EventType * eventP) {
	UInt16 formId;
	FormType * frmP;

	if (eventP->eType == frmLoadEvent) {
		/* Load the form resource. */
		formId = eventP->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);
 		switch (formId) {
			case MainForm:
				FrmSetEventHandler(frmP, MainFormHandleEvent);
				break;
		}
		return true;
	}

	return false;
}
static void AppEventLoop(void) {
	UInt16 error;
	EventType event;

	do {
		/* change timeout if you need periodic nilEvents */
		EvtGetEvent(&event, evtWaitForever);
		if (! SysHandleEvent(&event)) {
			if (! MenuHandleEvent(0, &event, &error)) {
				if (! AppHandleEvent(&event)) {
					FrmDispatchEvent(&event);
				}
			}
		}
	} while (event.eType != appStopEvent);
}
static Err AppStart(void) {
	return errNone;
}
static void AppStop(void) {
	FrmCloseAllForms();
}
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err error=errNone;
	switch (cmd)
	{
		case sysAppLaunchCmdNotify:
			switch (((SysNotifyParamType *)cmdPBP)->notifyType) {
				case sysNotifyMenuCmdBarOpenEvent:
					AddMyCommandBarItem();
					break;
				case appFileCreator:
					PerformDA();
					break;
				default:
					break;
			}
			break;
		case sysAppLaunchCmdSystemReset:
			AddMenu();
			break;
		case sysAppLaunchCmdNormalLaunch:
			error = AppStart();
			if (error)
				return error;

			/*
			 * start application by opening the main form
			 * and then entering the main event loop
			 */
			FrmGotoForm(MainForm);
			AppEventLoop();
			AppStop();
			break;
 	}
	return error;
}
